# -*- coding: utf-8 -*-
"""Скрипт для конвертирования фреймов в события в df файлах."""
# pylint: disable-msg=E0611

from argparse import ArgumentParser
from glob import glob
from os import makedirs
from os import path
import sys
import struct

from contextlib import closing
import dfparser
from multiprocess import Pool
from tqdm import tqdm

MAIN_DIR = path.abspath(path.join(path.dirname(__file__), '..'))
if MAIN_DIR not in sys.path:
    sys.path.append(MAIN_DIR)
del MAIN_DIR

from signal_utils.convert_utils import df_frames_to_events
from signal_utils.extract_utils import extract_amps_approx


def parse_args():
    """Парсинг параметров."""
    parser = ArgumentParser(description='Dataforge Lan10-12PCI frames to '
                            'events converter.')

    parser.add_argument('data_path', help='Path to root folder with data.')
    parser.add_argument('out_path', help='Path to root output folder.')
    parser.add_argument('wildcard',
                        help='Files wildcard relative to root folder.')

    return parser.parse_args()


if __name__ == "__main__":
    ARGS = parse_args()

    FILES_ABS_WILDCARD = path.join(ARGS.data_path, ARGS.wildcard)
    FILES = glob(FILES_ABS_WILDCARD, recursive=True)

    def process_file(data_file):
        """Обработка отдельного файла."""
        try:
            filepath_rel = path.relpath(data_file, ARGS.data_path)
            filepath_out = path.join(ARGS.out_path, filepath_rel)
            _, meta_real, data_real = dfparser.parse_from_file(data_file)
            meta_real_, data_real_ = df_frames_to_events(meta_real, data_real,
                                                         extract_amps_approx,
                                                         correct_time=True)

            if not path.exists(path.dirname(filepath_out)):
                makedirs(path.dirname(filepath_out))

            with open(filepath_out, "wb") as out_file:
                out_file.write(dfparser.create_message(meta_real_, data_real_))
        except struct.error:
            pass


    with closing(Pool()) as pool:
        with tqdm(total=len(FILES)) as pbar:
            for i, _ in tqdm(enumerate(pool.imap_unordered(process_file,
                                                           FILES))):
                pbar.update()



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 16 11:49:13 2017

@author: kapot
"""

import datetime
from os import path
from zipfile import ZipFile

import dfparser
import pandas as pd

gauth = None
drive = None

def __init_grive():
    global gauth
    global drive
    
    if not gauth:
        from pydrive.auth import GoogleAuth
        gauth = GoogleAuth()
        gauth.LocalWebserverAuth()
    
    if not drive:
        from pydrive.drive import GoogleDrive
        drive = GoogleDrive(gauth)
    

def get_points_drom_drive():
    __init_grive()
    points_raw = drive.ListFile({'q': "title contains '.rsb.zip'"}).GetList()

    times = [datetime.datetime.strptime(p["originalFilename"], \
             "%Y%m%d-%H%M%S.rsb.zip") for p in points_raw]

    pd.set_option('display.max_columns', 500)
    points = pd.DataFrame({'name': [p["originalFilename"] for p in points_raw],
                           'time': times,
                           'id': [p["id"] for p in points_raw],
                           'size': [int(p["fileSize"]) for p in points_raw]})
    return points


def load_dataset(id, name, size=0):
    
    """
      Скачать файл данных платы ЛАН10-12PCI с диска и открыть его парсером
      @id - идентификатор файла на гугл диске
      @name - исходное имя файла на гугл диске
      @size - размер файла в байтах на диске. Используется для проверки 
      скачанного файла
      @return - открытый датасет
      
    """
    
    __init_grive()
    if not(path.exists(name) and path.getsize(name) == size):
        file = drive.CreateFile({"id": id})
        file.GetContentFile(name)
    
    extracted_name = path.basename(name[:-4])
    
    if not path.exists(extracted_name):
        unpacked = ZipFile(name, "r")
        extracted_name = unpacked.extract(path.basename(name[:-4]))

    dataset =  dfparser.RshPackage(extracted_name)
    return dataset





"""Comparison of MADC and Lan10-12PCI full set spectrum.

Input:
- data - path to MADC data
- df_data_root - path to Lan10-12PCI root folder
- points_path - list for 14-17 kV points.

Output:
- Graph of full spectrum for MADC and Lan10-12PCI data.
- Graph of full spectrum ratio between MADC and Lan10-12PCI data.
  Formula: ratio[bin] = 1.0 - counts[bin]/counts_madc[bin]
"""
# pylint: disable-msg=R0914,R0915

from glob import glob
from os import path

import dfparser
import matplotlib.pyplot as plt
import numpy as np
import seaborn


def parse_madc_binary(data):
    """Extract amplitudes from madc binary data."""
    amps = np.zeros(len(data)//7, np.uint16)
    for amp_idx, idx in enumerate(range(0, len(data), 7)):
        amps[amp_idx] = np.frombuffer(data[idx:idx+2], np.uint16)[0]

    return amps


def main():
    """Execute main function."""
    df_madc_data_root = "/home/chernov/data_on_server_madc"
    df_data_root = "/home/chernov/data_processed"
    set_path = "2017_05/Fill_2/set_8"

    threshold_madc = 450
    threshold_madc_h = 3100
    threshold = threshold_madc*2.03
    threshold_h = threshold_madc_h*2.03

    points = sorted(glob(path.join(df_data_root, set_path, "p**")))
    points_madc = sorted(glob(path.join(df_madc_data_root, set_path, "p**")))

    counts_madc = {}

    for point in points_madc:
        _, meta, data = dfparser.parse_from_file(point)
        hv_val = int(meta['external_meta']['HV1_value'])

        if hv_val not in counts_madc:
            counts_madc[hv_val] = []

        amps = parse_madc_binary(data)

        filt = np.logical_and(amps > threshold_madc,
                              amps < threshold_madc_h)
        counts_madc[hv_val].append(amps[filt].size)

    counts = {}

    for point in points:
        _, meta, data = dfparser.parse_from_file(point)

        hv_val = int(meta['external_meta']['HV1_value'])

        time = meta['params']['b_size'] * meta['params']['events_num'] / \
            meta['params']['sample_freq']

        point_ds = dfparser.Point()
        point_ds.ParseFromString(data)

        amps = np.hstack([list(block.events.amplitudes)
                          for block in point_ds.channels[0].blocks])

        if hv_val not in counts:
            counts[hv_val] = []
        filt = np.logical_and(amps > threshold,
                              amps < threshold_h)
        counts[hv_val].append(float(amps[filt].size)/time*30.0)

    counts = {key: np.mean(counts[key]) for key in counts}
    counts_madc = {key: np.mean(counts_madc[key]) for key in counts_madc}

    counts_x = list(counts.keys())
    counts_y = [counts[key] for key in counts.keys()]

    counts_madc_x = list(counts_madc.keys())
    counts_madc_y = [counts_madc[key] for key in counts_madc.keys()]

    fig, axes = plt.subplots()
    fig.canvas.set_window_title("energy_spectrum_compare")
    fig.suptitle("CAMAC MADC vs. Lan10-12PCI energy spectrums \n"
                 "Set - %s" % (set_path))
    axes.plot(counts_x, counts_y, 'ro', label='Lan10-12PCI')
    axes.plot(counts_madc_x, counts_madc_y, 'bo', label='MADC')
    axes.set_xlabel("Voltage, V")
    axes.set_ylabel("Efficient counts")
    axes.legend()

    print(counts_x)
    print(counts_madc_x)
    # assert counts_x == counts_madc_x

    fig, axes = plt.subplots()
    fig.canvas.set_window_title("energy_spectrum_compare_ratio")
    fig.suptitle("CAMAC MADC vs. Lan10-12PCI energy spectrums ratio \n"
                 "Set - %s" % (set_path))
    axes.plot(counts_x, 1 - np.array(counts_y) / np.array(counts_madc_y),
              'ro', label='')
    axes.set_xlabel("Voltage, V")
    axes.set_ylabel("MADS vs Lan10_12PCI ratio")


if __name__ == "__main__":
    seaborn.set_context("poster")
    main()



"""Модуль для тестирования параметров алгоритмов."""

from datetime import datetime
from os import path

from dfparser import Point
import numpy as np
from tqdm import tqdm

from signal_utils.draw_utils import draw_metrics
from signal_utils.extract_utils import extract_frames
from signal_utils.generation_utils import generate_df


def _extract_real_frames(meta, data, block_params, frame_l=15, frame_r=25):
    """Выделение кадров реальных событий из сгенерированного df файла.

    @meta - Метаданные df файла.
    @data - Бинарные данные df файла.
    @block_params - Параметры событий.
    @frame_l - Размер сохраняемого кадра в бинах слева.
    @frame_r -Размер сохраняемого кадра в бинах справа.
    @retutn - Массив кадров.
    """
    bin_time = 1e+9 / meta['params']['sample_freq']
    times = block_params[1::2]

    point = Point()
    point.ParseFromString(data)

    frames = np.zeros((times.size, frame_l + frame_r), np.float32)
    for channel in point.channels:
        for block in channel.blocks:
            for event in block.frames:
                event_data = np.frombuffer(event.data, dtype=np.int16)
                ev_l_time = (block.time + event.time)
                ev_r_time = ev_l_time + len(event_data)*bin_time

                l_index = np.searchsorted(times, ev_l_time)
                r_index = np.searchsorted(times, ev_r_time)

                peaks = [int(round((times[idx] - ev_l_time)/bin_time)) for
                         idx in range(l_index, r_index)]

                frames[l_index:r_index] = extract_frames(event_data, peaks,
                                                         frame_l, frame_r)

    return np.vstack(frames)


def _calc_metrics(amps_real, pos_real,
                  amps_extracted, pos_extracted,
                  singles_extracted, max_pos_err):
    """Вычисление метрик.

    В процессе тестирования определяются:
        - время работы алгоритма
        - количество ложных срабатываний
        - количество пропущеных событий
        - отношение классифицированных наложенных событий к реальным
        (реальным наложенным событием является выделенное событие, которое
         соответсвует нескольким реальным событиям)

    Алгоритм расчета метрик:
        1. Задание соответсвия между реальными и выделенными событиями. Для
        этого для каждого события из массива реальных данных ищется ближайшее
        по положению событие из массива выделенных событий.
        2. Подсчет пропущенных событий. Если расстояние между реальным и
        восстановленным событием больше порога (max_pos_err) - событие
        помечается как пропущенное.
        3. Подсчет ложных срабатываний. Ложным срабатыванием считается
        событие из массива восстановленных данных у которого нет ни одного
        соответсвия с реальным событием.
        4. Подсчет реальных наложений. Реальным наложением является событие
        из восстановленного массива, которое соответсвуюет нескольким
        событиям из рельных данных.
        Реальные наложения затем сравниваются с классифицированными
        наложениями, полученными алгоритмом выделения.
        5. Вычисление ошибок восстановления ампитуд (как наложенных так и
        одинарных событий).

    @amps_real -- массив амплитуд реальных событий
    @pos_real -- массив положений реальных событий в наносекундах
    @amps_extracted -- массив амплитуд выделенных событий
    @pos_extracted -- массив амплитуд выделенных положений в наносекундах
    @singles_extracted -- массив классификаций выделенных событий по
    наложенности
    @max_pos_err -- максимальное допустимое отличие положений реального
    события и соответсвующего ему выделенного события в наносекундах

    @return -- Рассчитанные метрики. Пример метрики:
        {
            'amps_extracted': array([ 2463.82592773, ...,  5950.55712891],
                                    dtype=float32),
            'amps_real': array([ 1547.03479004, ...,963.0748291 ],
                               dtype=float32),
            'doubles_detected': array([], dtype=int32),
            'doubles_real': array([  33,  ..., 3719], dtype=int64),
            'false_negatives': array([   0, ..., 4025]),
            'false_positives': array([2593, 2742]),
            'pos_extracted': array([  6.14400000e+04, ..., 3.35321920e+08],
                                   dtype=float32),
            'pos_real': array([  1.50009424e+04, ..., 3.35438528e+08],
                              dtype=float32),
            'real_detected_transitions': array([-1, ..., -1], dtype=int64),
            'singles_extracted': array([ True, ...,  True], dtype=bool),
            'total_detected': 3860,
            'total_real': 4026
         }
    Здесь:
        doubles_detected -- индексы событий, классифицированных как
        наложенные
        doubles_real -- индексы реальных наложенных событий.
        false_negatives -- индексы неопределенных классификатором событий
        false_positives -- индексы ложно определенных событий
        real_detected_transitions -- Массив соответсвия между рельными и
        выделенными событиями. Если соответсвия нету - индекс
        прирванивается -1.
        total_detected -- общее количество выделенных событий
        total_real -- общее количество реальных событий

    """
    metrics = {}
    metrics["total_real"] = len(amps_real)
    metrics["total_detected"] = len(amps_extracted)

    STEP = 1000
    global idxs_raw
    idxs_raw = np.full(pos_real.shape, -1, np.int)
    for i in tqdm(range(0, len(pos_real), STEP), desc="indexing"):
        pos_real_bl = pos_real[i:i+STEP]
        min_ind = np.searchsorted(pos_extracted, pos_real_bl[0] - max_pos_err)
        max_ind = np.searchsorted(pos_extracted, pos_real_bl[-1] + max_pos_err)
        pos_extr_bl = pos_extracted[min_ind:max_ind]

        if len(pos_real_bl) and len(pos_extr_bl):
            dists_ = np.abs(np.subtract.outer(pos_extr_bl, pos_real_bl))

            vals, idxs_rev, counts = np.unique(dists_.argmin(0),
                                               return_inverse=True,
                                               return_counts=True)
            mults = np.where(counts > 1)[0]
            for idx in mults:
                cols = np.where(idxs_rev == idx)[0]
                argsort_idxs = cols[np.argsort(dists_[:, cols].min(axis=0))]
                for j in range(argsort_idxs.size):
                    point = dists_[:, argsort_idxs[j]].argmin()
                    dists_[point, argsort_idxs[j]] += 320*j

            idxs_raw[i: i + STEP] = dists_.argmin(0) + min_ind

    idxs_raw = np.hstack(idxs_raw)

    dists = np.abs(pos_real - pos_extracted[idxs_raw])
    idxs_raw[dists > max_pos_err] = -1

    single_idxs, counts = np.unique(idxs_raw, return_counts=True)

    metrics["real_detected_transitions"] = idxs_raw
    metrics["false_negatives"] = np.arange(len(idxs_raw))[idxs_raw == -1]
    metrics["false_positives"] = np.setdiff1d(np.arange(len(pos_extracted)),
                                              single_idxs)

    doubles_real = single_idxs[counts > 1]
    doubles_real = doubles_real[doubles_real != -1]
    metrics["doubles_real"] = doubles_real
    metrics["doubles_detected"] = np.arange(len(singles_extracted))[
        np.where(singles_extracted == False)]

    return metrics


def extract_frames_big_pos_err(metrics, err):
    """Извлечение кадров событий с большой ошибкой по времени.

    @metrics - Посчитанные метрики. @note test_on_df должа быть выполнена с
    флагом extr_frames=True.
    @err - Минимальная ошибка в нс.
    @retutn - Массив кадров.
    """
    idxs_raw = metrics["real_detected_transitions"]
    pos_err = metrics["pos_real"][idxs_raw != -1] - \
        metrics["pos_extracted"][idxs_raw[idxs_raw != -1]]
    pos_err = metrics["pos_real"][idxs_raw != -1] - \
        metrics["pos_extracted"][idxs_raw[idxs_raw != -1]]
    pos_err = np.abs(metrics["pos_real"][idxs_raw != -1] -
                     metrics["pos_extracted"][idxs_raw[idxs_raw != -1]])
    bad_frames = metrics["frames"][idxs_raw[idxs_raw != -1]][pos_err > 320]

    return bad_frames


def extract_frames_false_pos(metrics):
    """Извлечение кадров ложно положительных событий.

    @metrics - Посчитанные метрики. @note test_on_df должа быть выполнена с
    флагом extr_frames=True.
    @retutn - Массив кадров.
    """
    return metrics['frames'][metrics["false_positives"]]


def extract_false_neg():
    """Извлечение кадров ложно отрицательных событий.

    @metrics - Посчитанные метрики. @note test_on_df должа быть выполнена с
    флагом extr_frames=True.
    @retutn - Массив кадров.
    """
    return metrics['frames_real'][metrics["false_negatives"]]


def test_on_df(meta, data, block_params, algoritm_func, max_pos_err=3200,
               extr_frames=False, frame_l=15, frame_r=25):
    """Тестирование с использованием генерируемых df файлов.

    @note - алгоритм протестирован только на стандартной частоте оцифровки

    @meta - метаданные df файла
    @data - бинарные данные df файла
    @block_params - параметры событий
    @algoritm_func - функция выделения событий из данных.
    Входные аргументы функции:
        - data - данные блока (np.array)
        - start_time - начало блока в наносекундах
        - threshold - порог
        - sample_freq - частота оцифровки данных
    Выходные аргументы функции:
        (params, singles)
        - params - информация о событиях в формате [amp1, pos1, amp2, pos2,
                                                    ...] (np.array).
        Позиция должна быть задана в наносекундах от начала данных
          - singles - массив bool содержащий флаги наложенности событий. Для
          наложенных событий флаг должен иметь значение False, для одинарных
          событий  - True.
     В качестве примера функции выделения см
     signal_utils.extract_utils.extract_amps_approx2.
     @max_pos_err - макисмальная ошибка выделения положения пика (нс)
     @extr_frames - сохранение кадров событий в метрику
     @frame_l - размер сохраняемого кадра в бинах слева
     @frame_r - размер сохраняемого кадра в бинах справа

     @todo изменить фунции детектирования

    """
    threshold = meta['process_params']['threshold']
    sample_freq = meta['params']['sample_freq']

    point = Point()
    point.ParseFromString(data)

    start = datetime.now()
    frames = []
    amps = []
    times = []
    singles = []
    for channel in point.channels:
        for i, block in enumerate(channel.blocks):
            amps_block = []
            times_block = []
            singles_block = []
            for event in block.frames:
                ev_data = np.frombuffer(event.data, np.int16)
                params, singles_raw = algoritm_func(ev_data, event.time,
                                                    threshold, sample_freq)

                if extr_frames:
                    peaks = np.round((params[1::2] -
                                      event.time) * sample_freq / 1e+9)

                    frames_block = extract_frames(ev_data, peaks,
                                                  frame_l, frame_r)
                    frames.append(frames_block)

                amps_block.append(params[0::2])
                times_block.append(params[1::2])
                singles_block.append(singles_raw)

            amps_block = np.hstack(amps_block).astype(np.float16)
            times_block = np.hstack(times_block)
            times_block = times_block + block.time
            singles_block = np.hstack(singles_block).astype(np.bool)

            amps.append(amps_block)
            times.append(times_block)
            singles.append(singles_block)

    amps = np.hstack(amps)
    times = np.hstack(times)
    singles = np.hstack(singles)
    if extr_frames:
        frames = np.vstack(frames)

    end = datetime.now()
    delta = (end - start).total_seconds()

    amps_real = block_params[0::2]
    times_real = block_params[1::2]

    metrics = _calc_metrics(amps_real, times_real,
                            amps, times,
                            singles, max_pos_err)

    out = {"amps_real": amps_real,
           "pos_real": times_real,
           "amps_extracted": amps,
           "pos_extracted": times,
           "singles_extracted": singles,
           "time_elapsed": delta,
           **metrics}

    if extr_frames:
        out['frames'] = frames
        out['frames_real'] = _extract_real_frames(meta, data, block_params,
                                                  frame_l, frame_r)

    return out


def __test():
    import seaborn

    seaborn.set_context("poster")

    from extract_utils import extract_amps_approx
    from pylab import rcParams

    rcParams['figure.figsize'] = 10, 10

    dist_path = path.join(path.dirname(__file__), 'data/dist.dat')

    meta, data, block_params = generate_df(time=1, threshold=700,
                                           dist_file=dist_path, freq=40e3)

    global metrics
    metrics = test_on_df(meta, data, block_params, extract_amps_approx,
                         extr_frames=True)
    draw_metrics(metrics)


if __name__ == '__main__':
    __test()



# -*- coding: utf-8 -*-
"""Effective count rate comparison for real and generated point.

Input:
- dist_path - path to generation distribution
- filename - path to uprocessed dataforge point (14 Kv)

Output:
- Effective count rate for real and generated data for several time thresholds.
"""
# pylint: disable-msg=R0914,C0413
from os import path

import dfparser
import matplotlib.pyplot as plt
import numpy as np
import seaborn
from signal_utils.convert_utils import df_frames_to_events
from signal_utils.extract_utils import extract_amps_approx
from signal_utils.generation_utils import generate_df
from signal_utils.test_utils import _extract_real_frames


def prepare_point(time_s=5, freq=40e3, amp_thresh=700,
                  extract_func=extract_amps_approx,
                  dist_path=path.join(path.dirname(__file__),
                                      '../../signal_utils/data/dist.dat')):
    """Фильтрация заведомо недеткетируемых событий из точки."""
    meta, data, block_params = generate_df(time=time_s, threshold=amp_thresh,
                                           dist_file=dist_path, freq=freq)

    real_frames = _extract_real_frames(meta, data, block_params,
                                       frame_l=3, frame_r=5)

    amps_real = real_frames.max(axis=1)
    amps_detectable = amps_real[amps_real > amp_thresh]

    meta_, data_ = df_frames_to_events(meta, data, extract_func)
    meta_["detectable_events"] = amps_detectable.size

    return meta_, data_


def df_events_to_np(_, data):
    """Конвертация массива точек из формата df в numpy array."""
    point = dfparser.Point()
    point.ParseFromString(data)

    amps = []
    times = []

    channel = point.channels[0]
    for block in channel.blocks:
        amps.append(block.events.amplitudes)
        times.append((np.array(block.events.times) + block.time))

    amps = np.hstack(amps)
    times = np.hstack(times)

    return amps, times


def get_time_filtered_crs(meta, times, time_thresh_ns=6000):
    """Фильтрация по времени.

    Фильтрация близких друг к другу событий и поиск эффективной скорости
    счета.

    """
    deltas = times[1:] - times[:-1]

    filtered = np.hstack([[0, ], np.where(deltas > time_thresh_ns)[0] + 1])
    filtered_out = np.setdiff1d(np.arange(times.size), filtered)

    filtered_time_s = deltas[filtered_out - 1].sum() * np.double(1e-9)

    block_len_s = meta['params']['b_size'] / meta['params']['sample_freq']
    total_time_s = block_len_s * meta['params']['events_num']

    tau = np.double(total_time_s - filtered_time_s) / np.double(filtered.size) - \
        np.double(time_thresh_ns) * np.double(1e-9)
    count_rate = np.double(1) / tau

    return count_rate


def get_crs(meta, times, start=320 * 2, end=320 * 63, step=320 * 2):
    """Вычисление скоростей счета для разных отсечений по времени."""
    time_thrs = np.arange(start, end, step)
    crs = [get_time_filtered_crs(meta, times, time_thresh_ns=time_thr) for
           time_thr in time_thrs]
    return time_thrs, crs


def crs_compare_different_timesteps():
    """Сравнение графиков скоростей счета при разных шагах.

    Тест показывает причину возникновения "пилы" на графиках.
    """
    freq = 42e3
    amp_thresh = 750
    time_s = 1.0
    dist_path = path.join(path.dirname(__file__),
                          '../../signal_utils/data/dist.dat')

    from signal_utils.extract_utils import extract_simple_amps

    meta_gen, data_gen = prepare_point(time_s=time_s, freq=freq,
                                       amp_thresh=amp_thresh,
                                       extract_func=extract_simple_amps,
                                       dist_path=dist_path)

    _, times_gen = df_events_to_np(meta_gen, data_gen)

    time_thrs_640, crs_gen_640 = get_crs(meta_gen, times_gen)
    time_thrs_1000, crs_gen_1000 = get_crs(meta_gen, times_gen, step=1000)
    time_thrs_500, crs_gen_500 = get_crs(meta_gen, times_gen, step=400)

    fig, axes = plt.subplots()

    fig.canvas.set_window_title('cr_steps_compare')

    axes.set_title("Effective Count Rate / Time threshold")
    axes.set_xlabel("Time Threshold, ns")
    axes.set_ylabel("Effective Count Rate, Hz")

    axes.plot(time_thrs_640, crs_gen_640, label="Step = 640 ns")
    axes.plot(time_thrs_500, crs_gen_500, label="Step = 400 ns")
    axes.plot(time_thrs_1000, crs_gen_1000, label="Step = 1000 ns")

    axes.legend(loc=4)


def main():
    """Функция main.

    Сравнение эффективной скорости счета для реальных и сгенерированных данных.

    """
    crs_compare_different_timesteps()


if __name__ == "__main__":
    seaborn.set_context("poster")
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Drawing df point events time intervals spectrum script.
Point should be processed before drawing.
"""
from argparse import ArgumentParser

import dfparser  # Numass files parser
import matplotlib.pyplot as plt  # plotting library
import numpy as np
import seaborn as sns  # matplotlib grahs visual enchancer


def __parse_args():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('input', help='Input dataforge point.')
    parser.add_argument('-t', '--ampl-threshold', type=float, default=0,
                        help='Time difference lower threshold in ns (default '
                        '- 0).')
    parser.add_argument('-m', '--ampl-max', type=float, default=1e+5,
                        help='Time difference upper threshold in ns (default '
                        '- 0).')
    parser.add_argument('-b', '--bins', type=int, default=40,
                        help='Histogram bins number (default - 40).')
    return parser.parse_args()


def _main():
    # Parse arguments from command line
    args = __parse_args()

    # Read dataforge point
    _, _, data = dfparser.parse_from_file(args.input)

    # Parse Binary data
    point = dfparser.Point()
    point.ParseFromString(data)

    # Extract event times from each block
    times = []
    for channel in point.channels:
        for block in channel.blocks:
            times.append(np.array(block.events.times, np.uint64))

    # Combine times into one array
    times = np.hstack(times)

    # Calculate time differences
    diffs = times[1:] - times[:-1]

    # Calculate histogram
    hist, bins = np.histogram(diffs, bins=args.bins, range=(
        args.ampl_threshold, args.ampl_max))

    # Calculate bins centers
    bins_centers = (bins[:-1] + bins[1:]) / 2

    # Drawing graph
    _, axes = plt.subplots()
    axes.set_title(args.input)
    axes.set_xlabel("Time, ns")
    axes.set_ylabel("Counts")
    axes.step(bins_centers, hist, where='mid')
    plt.show()


if __name__ == "__main__":
    sns.set_context("poster")
    _main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Drawing events from raw dataforge point.
"""
from argparse import ArgumentParser

import dfparser  # Numass files parser
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns  # matplotlib grahs visual enchancer
from matplotlib.widgets import Slider


def __parse_args():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('input', help='Input raw dataforge point.')
    return parser.parse_args()


def _main():
    # Parse arguments from command line
    args = __parse_args()

    # Read dataforge point
    _, _, data = dfparser.parse_from_file(args.input)

    # Parse Binary data
    point = dfparser.Point()
    point.ParseFromString(data)

    # Extract raw frames from each block
    frames = []
    for channel in point.channels:
        for block in channel.blocks:
            for frame in block.frames:
                frames.append(np.frombuffer(frame.data, np.int16))

    # Draw graph
    fig, axes = plt.subplots()
    plt.subplots_adjust(bottom=0.25)
    ax_frame = plt.axes([0.1, 0.10, 0.8, 0.03])
    # Add slider on graph
    s_freq = Slider(ax_frame, 'Frame', 1, len(frames) - 1, valinit=0)

    cur_plot, = axes.plot(frames[0])

    def update(_):
        """Update frame in graph."""
        # Get current frame number from slider
        frame = int(round(s_freq.val))
        # Change frame data
        cur_plot.set_ydata(frames[frame])
        cur_plot.set_xdata(np.arange(len(frames[frame])))
        # Redraw plot
        fig.canvas.draw_idle()

    # Bind update function to slider
    s_freq.on_changed(update)
    plt.show()


if __name__ == "__main__":
    sns.set_context("poster")
    _main()



"""Dataforge point frames to events converting tool."""

from argparse import ArgumentParser
from os import path

import dfparser

from signal_utils.convert_utils import df_frames_to_events
from signal_utils.extract_utils import extract_amps_approx2


def _parse_args():
    parser = ArgumentParser(description='Dataforge point frames to '
                            'events converting tool.')

    parser.add_argument('input', help='Input dataforge file.')
    parser.add_argument('-o', '--output', default=None,
                        help='Output dataforge file. '
                        '(default - %filename%_extr.df')
    parser.add_argument('--frame-l', default=15,
                        help='Doubles left frame size. (default - 15)')
    parser.add_argument('--frame-r', default=25,
                        help='Doubles right frame size. (default - 25)')

    return parser.parse_args()


def _main():
    args = _parse_args()

    if args.output is None:
        output = path.join(path.dirname(args.input),
                           "%s_extr.df" % (path.splitext(args.input)[0]))
    else:
        output = args.output

    _, meta, data = dfparser.parse_from_file(args.input)
    meta_out, data_out = df_frames_to_events(meta, data, extract_amps_approx2,
                                             frame_l=args.frame_l,
                                             frame_r=args.frame_r)
    with open(output, "wb") as out_file:
        out_file.write(dfparser.create_message(meta_out, data_out))


if __name__ == "__main__":
    _main()



"""Misc utils for point processing."""

import collections

import numpy as np
from scipy.signal import argrelextrema
from tqdm import tqdm

from gdrive_utils import load_dataset


def extract_from_dataset(dataset, threshold=700, area_l=50, area_r=100):
    """Получение блоков из датасета [Desperated].

    @dataset - датасет
    @threshold - порог zero-suppression
    @area_l - левая область zero-suppression
    @area_r - правая область zero-suppression
    @return - вырезанные блоки

    """
    frames = []  # кадры из точки
    for i in range(dataset.params["events_num"]):
        try:
            frames.append(dataset.get_event(i)["data"])
        except Exception:
            break

    blocks = []  # вырезанные из кадра события
    for frame in frames:
        peaks = np.where(frame > threshold)[0]
        dists = peaks[1:] - peaks[:-1]
        gaps = np.append(np.array([0]), np.where(dists > area_r)[0] + 1)
        for gap in range(0, len(gaps) - 1):
            l_bord = peaks[gaps[gap]] - area_l
            r_bord = peaks[gaps[gap + 1] - 1] + area_r
            blocks.append(frame[l_bord: r_bord])
    return blocks


def extract_frames(ev_data, peaks, frame_l, frame_r):
    """Выделение событий из кадра по пикам.

    @ev_data - массив кадра
    @peaks - массив положений пиков в кадре в бинах
    @frame_l - размер обрезания события слева от пика
    @frame_r - размер обрезания события справа от пика
    @return - Двумерный массив событий.

    """
    shape = (len(peaks), frame_l + frame_r)
    frames_block = np.zeros(shape, np.int16)
    for j, peak in enumerate(peaks):
        if peak >= frame_l and peak < len(ev_data) - frame_r:
            frames_block[j] = ev_data[int(peak) - frame_l: int(peak) + frame_r]
    return frames_block


def calc_center(event, amp_arg, step=2):
    """Параболическая аппроксимация центра пика.

    @event - массив события
    @amp_arg - положение пика

    """
    x1 = amp_arg
    y1 = event[x1]
    x2 = x1 + step
    y2 = event[x2]
    x3 = x1 - step
    y3 = event[x3]

    a = (y3 - (x3*(y2-y1) + x2*y1 - x1*y2)/(x2 - x1))/ \
      (x3*(x3 - x1 - x2) + x1*x2)
    b = (y2 - y1)/(x2 - x1) - a*(x1 + x2)
    c = (x2*y1 - x1*y2)/(x2 - x1) + a*x1*x2

    x0 = -b/(2*a)
    y0 = a*x0**2 + b*x0 + c

    return y0, x0


def get_peaks(event, threshold):
    """Выделение пиков из блока.

    Условия отбора пика:
    - пик должен быть больше порогового значения
    - пик должен быть локальным экстремумом

    @event - массив события
    @threshold - порог

    """
    extremas = argrelextrema(event, np.greater_equal)[0]
    points = extremas[event[extremas] >= threshold]

    if len(points) >= 2:
        planes = np.where(points[1:] - points[:-1] <= 3)[0] + 1
        points = np.delete(points, planes)

    return points


def get_blocks(points, idxs, threshold=700, area_l=50, area_r=100):
    """
      [Desperated]
      Выделение из файлов google drive блоков алгоритмом zero-suppression
      @points - таблица с точками
      @idxs - индекс или массив индексов интересующих точек
      @threshold - порог zero-suppression
      @area_l - левая область zero-suppression
      @area_r - правая область zero-suppression
      @return - вырезанные блоки

    """
    blocks = []

    def add_blocks(blocks, idx):
        header = points.as_matrix()[idx]
        dataset = load_dataset(header[0], header[1], header[2])
        blocks += extract_from_dataset(dataset, threshold, area_l, area_r)

    if isinstance(idxs, collections.Iterable):
        for i in tqdm(idxs):
            add_blocks(blocks, i)
    else:
        add_blocks(blocks, idxs)

    return blocks


def get_bin_sec(points, idx):
    """
      Получение из файла google drive частоты оцифровки [Desperated].
      
      @points - таблица с точками
      @idx - индекс точки, из которой берется частота оцифровки

    """
    header = points.as_matrix()[idx]
    dataset = load_dataset(header[0], header[1], header[2])
    return dataset.params["sample_freq"]**-1



from os import path

import numpy as np
from scipy.optimize import curve_fit

from generation_utils import gen_multiple
from generation_utils import gen_signal
from process_utils import calc_center
from process_utils import extract_frames
from process_utils import get_peaks


def extract_fit_all(data, start_time, threshold, sample_freq,
                    pos_step=2.0, amp_step=500.0):
    """
      Выделение событий с помощью фитирования всех событий одновременно.
      Классификация наложений производится полносвязной нейронной сетью.

      Алгоритм требует высокой производительности, но дает наиболее точный
      результат.

      Алгоритм:
          1. Выделение локальных максимумов в кадре выше порога. Локальные
          максимумы считаются предварительными пиками.
          2. Классификация наложенных событий с помощью НС по предварительным
          пикам.
          3. Одновременное фитирование всех пиков. В качестве начальных данных
          берутся предварительные пики и соответсвующие им амплитуды. Диапазоны
          фитирования задаются значениями pos_step и amp_step.

    """
    if not extract_fit_all.model:
        if not 'load_model' in locals():
            from keras.models import load_model
        extract_fit_all.model = load_model(path.join(path.dirname(__file__),
                                           "data/mlp_classifier.h5"))

    peaks = get_peaks(data, threshold)

    if not len(peaks):
        return np.array([]), np.array([])

    prep_peaks = np.vstack([data[peak - extract_fit_all.l_off:
                                 peak + extract_fit_all.r_off]
                           for peak in peaks])
    singles = extract_fit_all.model.predict(prep_peaks/extract_fit_all.x_max)\
                           .argmax(axis=1).astype(np.bool)

    params = extract_events_fit_all(data, threshold, amp_step, pos_step)

    params[1::2] = ((params[1::2]/sample_freq)*1e+9) + start_time

    return params, singles

extract_fit_all.model = None
extract_fit_all.frame_len = 50
extract_fit_all.l_off = 14
extract_fit_all.r_off = extract_fit_all.frame_len - extract_fit_all.l_off
extract_fit_all.x_min = -32768
extract_fit_all.x_max = 32768


def extract_simple_amps(data, start_time, threshold, sample_freq):
    """
      Выделение событий поиском локальных максимумов выше порога.

      Алгоритм является наиболее быстрым, однако он не учитывает форму сигнала,
      что приводит к систематической ошибке выделения амлитуд близких событий
      (событие попадает на хвост другого события).

    """
    peaks = get_peaks(data, threshold)

    coeff = np.float128(1e+9/sample_freq)

    params = np.zeros(len(peaks)*2, np.float128)
    params[0::2] = data[peaks]
    params[1::2] = (np.float128(peaks)*coeff) + start_time
    singles = np.ones(peaks.shape, np.bool)

    return params, singles


def extract_amps_approx(data, start_time, threshold, sample_freq):
    """
      Последовательное выделение событий из блока с вычитанием предыдущих
      событий.

      Алгоритм имеет меньшую скорость работы по сравнению с
      extract_simple_amps, однако учитывет форму сигнала.

      Алгоритм:
          1. Выделение локальных максимумов в кадре выше порога.
          2. Последовательная обработка пиков:
              1. Сохранение амплитуды и положения текущего пика.
              2. Вычитание из данных формы сигнала, соответсвующей выделенным
              амплитуде и положению текущего события.
              3. Переход к следующему событию.

    """
    data = data.copy().astype(np.float32)
    peaks = get_peaks(data, threshold)

    params = np.zeros(len(peaks)*2, np.float32)
    params[1::2] = ((peaks/sample_freq)*1e+9) + start_time

    x = np.linspace(0, data.size/sample_freq, data.size)

    for i in range(len(peaks)):
        peak = peaks[i]
        amp = data[peak]
        params[i*2] = amp
        data -= gen_multiple(x, amp, peak/sample_freq)

    singles = np.ones(peaks.shape, np.bool)

    return params, singles


def extract_amps_approx2(data, start_time, threshold, sample_freq,
                         classify=False, frame_l=15, frame_r=25):
    """
      Последовательное выделение событий из блока с вычитанием предыдущих
      событий.

      Ускоренный вариант extract_amps_approx. Вместо вычета события из всего
      кадра, событие вычитается только из пиков.

      Примерно на 30% медленее extract_simple_amps.

      @todo: Разобраться с сохранением времени в наносекундах
    """

    peaks = get_peaks(data, threshold)
    if classify:
        if not extract_amps_approx2.bst:
            if not 'load_model' in locals():
                global xgb
                xgb = __import__('xgboost')
            extract_amps_approx2.bst = xgb.Booster()
            model_path = path.join(path.dirname(__file__), 'data/xgb.dat')
            extract_amps_approx2.bst.load_model(model_path)

        frames = extract_frames(data.copy(), peaks, frame_l, frame_r)
        preds = extract_amps_approx2.bst.predict(xgb.DMatrix(frames))
        singles = preds > 0.97
    else:
        singles = np.ones(peaks.shape, np.bool)

    params = np.zeros(len(peaks)*2, np.float32)
    params[0::2] = data[peaks]
    params[1::2] = ((peaks/sample_freq)*1e+9) + start_time

    for i, peak in enumerate(peaks[:-1]):
        params[(i + 1)*2::2] -= gen_signal(peaks[i + 1:], params[i*2], peak)

    return params, singles
extract_amps_approx2.bst = None


def extract_amps_approx3(data, start_time, threshold, sample_freq):
    """
      Последовательное выделение событий из блока с вычитанием предыдущих
      событий.
    """
    data = data.copy().astype(np.float64)
    x = np.linspace(0, data.size/sample_freq, data.size)

    params = []
    while True:
        peaks = get_peaks(data, threshold)
        if not len(peaks):
            break

        peak = peaks[0]

        params.append(data[peak])
        params.append(((peak/sample_freq)*1e+9) + start_time)

        data -= gen_signal(x, data[peak], peak/sample_freq)

    params = np.array(params)
    singles = np.ones((params.shape[0]//2), np.bool)

    return params, singles


def extract_amps_front_fit(data, start_time, threshold, sample_freq,
                           l_step=2, r_step=5):
    """Выделение событий с помощью последовательного фитирования фронта.


    @l_step - Количество точек слева от порогового бина, по которым будет
    проведено фитирование.
    @r_step - Количество точек справа от порогового бина, по которым будет
    проведено фитирование.
    """
    params = []
    while True:
        data, popt, pcov = extract_event(data, threshold, l_step=l_step,
                                         r_step=r_step, freq=1.0/sample_freq)
        if popt is None:
            break
        else:
            params.append(popt)

    params = np.hstack(params)
    singles = np.ones(params.shape[0]//2, np.bool)


    params[1::2] = params[1::2]*1e+9 + start_time
    return params, singles


def extract_event(data, threshold, l_step, r_step, freq, def_amp=2000):
    """Выделение первого события в кадре с помощью фитирования по фронту.

    @data - Кадр события.
    @threshold - Порог.
    @def_amp - Амплитуда по-умолчанию при фитировании.
    @l_step - Количество точек слева от порогового бина, по которым будет
    проведено фитирование.
    @r_step - Количество точек справа от порогового бина, по которым будет
    проведено фитирование.
    @freq - Частота оцифровки.
    @return data, popt, pcov, где:
        data - Кадр, из которого вычтено определившееся событие.
        popt - [amp, pos] подобранные при фитировании амплитуда и положение
        события
        pcov - параметры сходимости (см scipy.optimize.curve_fit)
    """
    above_thr = np.where(data > threshold)[0]
    if not above_thr.size:
        return data, None, None

    bord = above_thr[0]

    l_ind = max(0, bord - l_step)
    r_ind = min(data.size, bord + r_step)

    part = data[l_ind:r_ind]

    x = np.linspace(0, part.shape[0]*freq, part.shape[0])

    p0 = [def_amp, part.argmax()*freq]
    bounds = ([500, part.argmax()*freq - 3.2e-06],
              [5700, part.argmax()*freq + 3.2e-06])

    popt, pcov = curve_fit(gen_signal, x, part,
                           p0=p0, bounds=bounds)

    x_big = np.linspace(0, data.shape[0]*freq, data.shape[0])

    return data - gen_signal(x_big - l_ind*freq, *popt), \
           popt + [0, l_ind*freq], pcov


def extract_events_fit(ev, threshold):
    """
      Последовательное выделение событий из блока

      В блоке последовательно фитируются событие. Функция следующего события
      складывается с фунциями уже определенных ранее событий.

      Данный метод работает потенциально быстрее и стабильнее по сравнению с
      одновременным фитированием всех событий в блоке, выдает большую ошибку
      при выделении близко наложенных событий (событие накладывается на хвост
      предыдущего события).

      @ev - массив события
      @threshold - порог
      @return параметры событий в формате [amp1, pos1, amp2, pos2, ...]

    """
    points = get_peaks(ev, threshold)
    values = np.array([], np.float32)

    for i, point in enumerate(points):
        y0, x0 = calc_center(ev, point)

        full_func = lambda x, a, p: gen_multiple(x, a, p, *values)

        popt, pcov = curve_fit(full_func, np.arange(len(ev)), ev, p0=[y0, x0])
        values = np.hstack([values, popt])

    return values


def extract_events_fit_all(ev, threshold, pos_step=2.0, amp_step=500.0):
    """
      Последовательное выделение событий из блока
      Все события фитируются одновременно.

      @ev - массив события
      @threshold - порог
      @pos_step - максимальное отклонение по положению при фитировании
      @amp_step - максимальное положительное отклонение по амплитуде при
      фитировании. Нижняя граница всегда равна 0.
      @return параметры событий в формате [amp1, pos1, amp2, pos2, ...]

    """
    points = get_peaks(ev, threshold)

    values = np.zeros(len(points)*2)
    values[0::2] = ev[points]
    values[1::2] = points

    upper_bounds = values.copy()
    upper_bounds[1::2] = upper_bounds[1::2] + pos_step
    upper_bounds[0::2] = upper_bounds[0::2] + amp_step

    lower_bounds = values.copy()
    lower_bounds[1::2] = lower_bounds[1::2] - pos_step
    lower_bounds[0::2] = 0

    full_func = lambda x, *values: gen_multiple(x, *values)
    popt, pcov = curve_fit(full_func, np.arange(len(ev)), ev, p0=list(values))

    return popt


def extract_algo(data, threshold=700):
    """
      Выделенние события из блока. Способ, предложенный Пантуевым В.С.
      @data - массив кадра
      @threshold - порог
      @return - индекс первого бина, превысившего порог,
      индекс перегиба, индекс первого отрицательного бина после перегиба

    """
    deriv = data[1:] - data[:-1]
    first_greater = np.argmax(data >= threshold)
    extremum = first_greater + np.argmax(deriv[first_greater-1:] < 0) - 1
    first_negative = first_greater + np.argmax(data[first_greater:] < 0)
    return first_greater, extremum, first_negative


def extract_events(data, threshold=700):
    """
      Выделенние нескольких событий из блока.
      Способ, предложенный Пантуевым В.С.
      @data - массив кадра
      @threshold - порог
      @return - [индекс первого бина, превысившего порог,
      индекс перегиба, индекс первого отрицательного бина после перегиба]

    """
    events = []
    offset = 0
    while(True):
        left, center, right = extract_algo(data[offset:], threshold)
        if not left:
            break
        left += offset
        center += offset
        r_buf = right
        right += offset
        offset += r_buf
        events.append([left, center, right])

    return events



# -*- coding: utf-8 -*-
"""Расчет гистограммы амплитуд отброшенных при временной фильтрации событий.

На входе должна быть одна обработанная точка в формате .df.
"""

from os import path

import dfparser
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from test_time_filtering import df_events_to_np

TIME_FILTER_THRESH = 6000
AMPL_THRESH = 496
AMPL_MAX = 4016
BINS = 55
DATA_ROOT = "/home/chernov/data/lan10_processed/"
POINT_PATH = "2017_11/Fill_3/set_1/p81(30s)(HV1=14950).df"


def filter_bad_events(meta, times, events, time_thresh_ns=6000):
    """Фильтрация и вывод массива событий, не прошедших временную фильтрацию.

    @meta - метаданные точки
    @times - массив времен точки
    @events - массив времен точки
    @time_thresh_ns - порог по времени в нс

    @return - массив отфильтрованных плохих событий

    """
    deltas = times[1:] - times[:-1]

    filtered = np.hstack([[0, ], np.where(deltas > time_thresh_ns)[0] + 1])
    filtered_out = np.setdiff1d(np.arange(times.size), filtered)

    return events[filtered_out]


def __main():
    _, meta, data = dfparser.parse_from_file(path.join(DATA_ROOT, POINT_PATH))
    amps, times = df_events_to_np(_, data)

    amps_filtered = filter_bad_events(meta, times, amps, TIME_FILTER_THRESH)

    _, hist_ax = plt.subplots()
    hist_ax.set_title("Rejected events histogramm")
    hist_ax.set_xlabel("Amplitude, ch")
    hist_ax.set_ylabel("Events, num")

    hist, bins = np.histogram(amps_filtered, range=(
        AMPL_THRESH, AMPL_MAX), bins=BINS)
    hist_ax.step((bins[1:] + bins[:-1]) / 2, hist,
                 where="mid", label="Rejected")
    hist, bins = np.histogram(amps, range=(AMPL_THRESH, AMPL_MAX), bins=BINS)
    hist_ax.step((bins[1:] + bins[:-1]) / 2, hist, where="mid", label="All")


if __name__ == "__main__":
    sns.set_context("poster")
    DEF_PALLETE = sns.color_palette()
    __main()



"""Signal utils init script."""
from pkg_resources import get_distribution

from . import (convert_utils, draw_utils, extract_utils, gdrive_utils,
               generation_utils, test_utils)

__version__ = get_distribution('signal-utils').version



"""Comparison of MADC and Lan10-12PCI acquired spectrum for several points.

Input:
- MADC spectrum - https://drive.google.com/open?id=0B0Ux_fvsLMdAT0cxSm1qZzFKcDQ
- Lan10-12PCI spectrum - set_43, 14 - 17 kV

Variables:
data - path to MADC data
df_data_root - path to Lan10-12PCI root folder
points_path - list for 14-17 kV points.


Output:
- points spectrum diffrerence graphs for every point.
"""
# pylint: disable-msg=R0914,W0640

from os import path

import dfparser
import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
import seaborn


def main():
    """Compare Lan10-12PCI points energy spectrum for several points."""
    seaborn.set_context("poster")

    data = np.genfromtxt("/home/chernov/Downloads/set_43_detector.out")

    x_data = data[:, 0]

    y_points = data[:, 1:].transpose()
    y_points = (y_points.T / np.max(y_points, axis=1)).T

    df_data_root = "/home/chernov/data_processed"

    points_path = ["2017_05/Fill_3/set_43/p0(30s)(HV1=16000).df",
                   "2017_05/Fill_3/set_43/p36(30s)(HV1=17000).df",
                   "2017_05/Fill_3/set_43/p80(30s)(HV1=15000).df",
                   "2017_05/Fill_3/set_43/p102(30s)(HV1=14000).df"]

    bins = 500
    range_ = (0, 8000)

    for idx, point_rel in enumerate(points_path):
        _, _, data = dfparser.parse_from_file(path.join(df_data_root,
                                                        point_rel))

        point = dfparser.Point()
        point.ParseFromString(data)

        amps = np.hstack([list(block.events.amplitudes)
                          for block in point.channels[0].blocks])

        hist, x_point = np.histogram(amps, bins=bins, range=range_)
        hist = hist / np.max(hist[np.where(x_point > 3000)[0][0]:])

        func = interp1d(x_point[1:] + x_point[:-1], hist,
                        bounds_error=False, fill_value=0)

        func_mod = lambda x, a, b, c: c*func(a*x + b)

        x_peak = np.where(np.logical_and(x_data > 1000, x_data < 1600))
        popt, _ = curve_fit(func_mod, x_data[x_peak], y_points[idx][x_peak],
                            p0=[3.68, 700, 1])

        fig, axes = plt.subplots()
        fig.canvas.set_window_title(point_rel)
        fig.suptitle("CAMAC MADC vs. Lan10-12PCI spectrums")
        axes.set_title("File - %s. \nOptimized parameters: a=%s, b=%s, c=%s" %
                       (point_rel, *np.round(popt, 2)))
        axes.set_xlabel("Bins, ch")
        axes.set_xlim(0, 2000)
        # axes.set_yscale("log", nonposx='clip')
        x_interp = np.linspace(0, 2000, 500)
        axes.plot(x_interp, func_mod(x_interp, *popt), label="Lan10-12PCI")
        axes.plot(x_data, y_points[idx], label="CAMAC MADC")
        axes.legend()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb  7 11:32:51 2018

@author: chernov
"""
import csv
import io
import re
from argparse import ArgumentParser, ArgumentTypeError
from glob import glob
from os import path

import dateutil
import dfparser  # Numass files parser
import matplotlib.pyplot as plt  # plotting library
import seaborn as sns  # matplotlib grahs visual enchancer
from dateutil.parser import parse as dateparse  # timestamp parser


def parse_args():
    """Parse arguments from command line."""
    def mark(val):
        """Mark parser type.

        Type represents time mark with label in format: LABEL,TIME_ISO
        """
        try:
            label, timestamp = val.split(',')
            return label, dateparse(timestamp)
        except ValueError:
            raise ArgumentTypeError("Coordinates must be LABEL,TIMESTAMP_ISO")

    parser = ArgumentParser(description='Slow control graph drawer')
    parser.add_argument('input', help='Input dataforge file.')
    parser.add_argument('-t', '--title', type=str, default=None,
                        help='Graph title (default - filename).'
                        'The flag can be used repeatedly in command.')
    parser.add_argument('--x-scale', default=None,
                        choices=['linear', 'log', 'logit', 'symlog'],
                        help='Axis x scale (default - linear).')
    parser.add_argument('--y-scale', default=None,
                        choices=['linear', 'log', 'logit', 'symlog'],
                        help='Axis y scale (default - linear).')
    parser.add_argument('-e', '--exclude', action='append', default=[],
                        help='Exclude graphs names. '
                        'The flag can be used repeatedly in command.')
    parser.add_argument('-m', '--mark', action='append', type=mark, default=[],
                        help='Add mark with label on graph in '
                        'LABEL,TIMESTAMP_ISO format '
                        '(ex: -m mark1,2017-11-19T09:44:11Z). '
                        'The flag can be used repeatedly in command.')
    return parser.parse_args()


def _main():
    args = parse_args()
    # Read dataforge point

    # Creating graph
    _, axes = plt.subplots()

    for idx_f, input_file in enumerate(glob(args.input)):
        header, _, data = dfparser.parse_from_file(input_file)
        # Read binary data manually due to machine header error
        # Dont need for good data
        header_len = dfparser.type_codes.ENVELOPE_HEADER_CODES[
            header["type"]]["header_len"]

        with (open(input_file, 'rb')) as raw_file:
            raw_file.seek(header_len + header['meta_len'])
            data = raw_file.read()

        # Prettify data. Remove double spaces
        data_tabed = re.sub(b'[ \t]+', b'\t', data)
        # Remove icorrect column name '#f\t' from data
        data_tabed = data_tabed[3:]

        # Parse data as TSV format
        tsv = csv.DictReader(io.StringIO(
            data_tabed.decode()), delimiter='\t')

        # Extracting values from parsed TSV
        graphs = {}
        for row in tsv:
            for key in row.keys():
                if key != 'timestamp':
                    if row[key] != '@null':  # Filter null values
                        if key not in graphs:
                            graphs[key] = {'x': [], 'y': []}

                        # Append timestamp to point
                        graphs[key]['x'].append(
                            dateutil.parser.parse(row['timestamp']))
                        # Append value to point
                        graphs[key]['y'].append(
                            float(row[key]))

        # Plotting graphs
        for idx, graph in enumerate(graphs.keys()):
            if graph not in args.exclude:  # Filter excluded data
                label = None
                if idx_f == 1:
                    label = graph
                axes.plot(
                    graphs[graph]['x'], graphs[graph]['y'],
                    color=sns.color_palette()[idx], label=label)

    for text, x_coord in args.mark:
        axes.annotate(text, xy=(x_coord, 0), xytext=(x_coord, 1e2),
                      arrowprops=dict(facecolor='black', shrink=0.05))

    # Applying parameters
    if args.title:
        axes.set_title(args.title)
    else:
        axes.set_title(path.basename(args.input))
    if args.x_scale:
        axes.set_xscale(args.x_scale)
    if args.y_scale:
        axes.set_yscale(args.y_scale)
    axes.legend()
    plt.show()


if __name__ == "__main__":
    sns.set_context("poster")
    _main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 18 11:16:55 2018.

@author: chernov

Скрипт получает симулированную форму спектра двойных наложений с помощью
генерации только двойных событий и последущей обработки созданного файла.

"""
# pylint: disable-msg=E1101,R0914

from os import path

import dfparser
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from signal_utils.convert_utils import df_frames_to_events
from signal_utils.extract_utils import extract_amps_approx2
from signal_utils.generation_utils import generate_df

TIME_SEC = 30
HIST_DIFF_RANGE = (0, 320 * 10)
HIST_DIFF_BINS_N = 40
HIST_EVS_RANGE = (0, 4096)
HIST_EVS_BINS_N = 64


def gen_doubles(ev_num, bl_time, max_spread=7 * 320e-9):
    """Функция распределения двойных событий.

    Алгоритм может создать и тройные наложения.

    @ev_num - количество событий
    @bl_time - время блока в секундах
    @max_spread - максимальное расстояние между сдвоенными событиями
    """
    times = np.zeros(ev_num + ev_num % 2, np.float128)
    times[0::2] = np.sort(np.random.uniform(
        0, bl_time, (ev_num + ev_num % 2) // 2))
    times[1::2] = times[0::2] + \
        np.random.uniform(0, max_spread, (ev_num + ev_num % 2) // 2)
    return times[:ev_num]


def __main():
    dist_file = path.abspath(path.join(path.dirname(
        __file__), "../../signal_utils/data/dist.dat"))
    meta, data, answer = generate_df(
        time=TIME_SEC, dist_file=dist_file,
        dist_time_func=gen_doubles, freq=10e3)
    ev_times = answer[1::2]
    ev_time_diffs = ev_times[1:] - ev_times[:-1]

    _, dist_hist_ax = plt.subplots()
    dist_hist_ax.set_title("Histogram of distances between generated events")
    dist_hist_ax.set_xlabel("Distance, ns")
    dist_hist_ax.set_ylabel("Events number")
    hist, bins = np.histogram(ev_time_diffs, bins=HIST_DIFF_BINS_N,
                              range=HIST_DIFF_RANGE)
    dist_hist_ax.step((bins[1:] + bins[:-1]) / 2, hist)

    _, data_ext = df_frames_to_events(meta, data, extract_amps_approx2)
    point = dfparser.Point()
    point.ParseFromString(data_ext)
    evs_recovered = []
    for channel in point.channels:
        for block in channel.blocks:
            evs_recovered.append(np.array(block.events.amplitudes))
    evs_recovered = np.hstack(evs_recovered)

    _, ampl_hist_ax = plt.subplots()
    ampl_hist_ax.set_title("Events amplutides histogram")
    ampl_hist_ax.set_xlabel("Amplitude, ch")
    ampl_hist_ax.set_ylabel("Events number")
    hist, bins = np.histogram(answer[0::2], bins=HIST_EVS_BINS_N,
                              range=HIST_EVS_RANGE)
    ampl_hist_ax.step((bins[1:] + bins[:-1]) / 2, hist, where='mid',
                      label="Generated events")
    hist, bins = np.histogram(evs_recovered, bins=HIST_EVS_BINS_N,
                              range=HIST_EVS_RANGE)
    ampl_hist_ax.step((bins[1:] + bins[:-1]) / 2, hist, where='mid',
                      label="Reconstructed events")
    ampl_hist_ax.legend()


if __name__ == "__main__":
    sns.set_context("poster")
    DEF_PALLETE = sns.color_palette()
    __main()



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 16 11:53:00 2017

@author: kapot
"""

import numpy as np
plt = None
plugins = None

def import_graphics():
    global plt
    global plugins
    if not plt:
        plt = __import__('matplotlib').pyplot
    if not plugins:
        plugins = __import__('mpld3').plugins

class graphics(object):

    def __init__(self, f):
        self.f = f

    def __call__(self, *args, **kwargs):
        import_graphics()
        self.f(*args, **kwargs)
    

@graphics
def draw_event(data, num, fig, ax):
    """
      Отрисовка первого события из блока. Размер события задан 128 бинам.
      @data - событие
      @num - номер события
      @fig - график
      @ax - оси
      
    """
    data = data[0:128]
    line = ax.plot(data)
    tooltip = plugins.LineLabelTooltip(line[0], label='event: %s'%(num))
    plugins.connect(fig, tooltip)


@graphics
def plot_event(data, num, event, ax, threshold=700, bin_sec=3.2e-07):
    """
      Отрисовка события и его пареметров
      @data - событие
      @num - номер события
      @event - событие в формате [индекс первого бина, превысившего порог,
      индекс перегиба, индекс первого отрицательного бина после перегиба]
      @ax - оси, на которые будут открисованы события
      @threshold - порог
      @bin_sec - коэфициент перевода бинов в секунды
      
    """
    y_range = (np.min(data), np.max(data))

    first_greater, extremum, first_negative = event

    lenght = first_negative - extremum
    ax.set_title("ampl: %s, size: %s (%s s)"%(data[extremum], 
                 lenght, lenght*bin_sec))

    ax.plot(data, label="event")
    ax.plot((0, len(data)), (threshold, threshold), "--", 
            color="green", label="first greater")
    ax.plot((first_greater, first_greater), y_range, "--", 
            color="orange", label="first greater")
    ax.plot((extremum, extremum), y_range, "--", 
            color="red", label="extremum")
    ax.plot((first_negative, first_negative), y_range, "--", 
            color="#00AAFF", label="first negative")

    ax.grid(True, alpha=0.3)
    ax.set_xlabel('time')
    ax.set_ylabel('ampl')
    ax.legend();


@graphics
def plot_multiple_events(data, events, ax, threshold=700, bin_sec=3.2e-07):
    """
      Отрисовка события и его пареметров
      @data - массив кадра
      @ax - оси, на которые будут открисованы события
      @events - события в формате [[индекс первого бина, превысившего порог,
      индекс перегиба, индекс первого отрицательного бина после перегиба],...]
      @threshold - порог
      @bin_sec - коэфициент перевода бинов в секунды
      
    """
    y_range = (np.min(data), np.max(data))

    ax.plot(data)
    ax.plot((0, len(data)), (threshold, threshold), "--", color="green")

    for ev in events:
        ax.plot((ev[0], ev[0]), y_range, "--", color="orange")
        ax.plot((ev[1], ev[1]), y_range, "--", color="red")
        ax.plot((ev[2], ev[2]), y_range, "--", color="#00AAFF")

    ax.grid(True, alpha=0.3)
    ax.set_xlabel('time')
    ax.set_ylabel('ampl')


@graphics
def draw_metrics(metrics: dict):
    """
      Отрисовка метрик
      
      Выводятся:
          - гистограмма расстояний между пиками в наложенных событиях
          - гистограмма ошибок по амплитудам
          - гистограммы реального и восстановленного спектров
          - гистограмма реальных амплитуд пропущенных событий (без шумов)
      
      @metrics -- Метрики. Cм. вывод signal_utils.test_utils._calc_metrics
      
    """
    print("Summary: ")
    print("Algoritm time: %s s"%(metrics["time_elapsed"]))
    print("Total events: %s"%(metrics["total_real"]))
    print("Detected events: %s"%(metrics["total_detected"]))
    print("Positives/negatives: ")
    print("False negatives %s"%(len(metrics["false_negatives"])))
    print("False positives %s"%(len(metrics["false_positives"])))

    idxs_raw = metrics["real_detected_transitions"]

    def get_dbl_amps_dists(idxs):

        dists, amps = [], []

        for idx in idxs:
            doubles_pos = metrics["pos_real"][idxs_raw == idx]
            doubles_amps = metrics["amps_real"][idxs_raw == idx]
            dists.append(doubles_pos[1:] - doubles_pos[:-1])
            amps.append(np.log(doubles_amps[1:], 
                               doubles_amps[:-1]))

        if len(dists):
            dists = np.hstack(dists)
            amps = np.hstack(amps)
            return amps, dists
        else:
            return np.array([]), np.array([])

    dbl_amps_real, dbl_dists_real = get_dbl_amps_dists(metrics["doubles_real"])
    dbl_amps_det, \
    dbl_dists_det = get_dbl_amps_dists(metrics["doubles_detected"])

    fig, ax = plt.subplots(4, 2)
    hist, bins = np.histogram(dbl_dists_real, 40)
    hist_det, bins = np.histogram(dbl_dists_det, bins=bins)
    ax[2][0].set_title("real doubles")
    ax[2][0].hist2d(dbl_dists_real, dbl_amps_real)
    ax[2][1].set_title("detected doubles")
    ax[2][1].hist2d(dbl_dists_det, dbl_amps_det)
    ax[1][1].set_title("real/detected doubles dists cumsums")
    ax[1][1].plot(bins[1:], np.cumsum(hist), label='real')
    ax[1][1].plot(bins[1:], np.cumsum(hist_det), label='detected')
    ax[1][1].legend()

    print("%s detected \n"\
          "%s real \n"\
          "%s intersection"%(len(metrics["doubles_detected"]),
                             len(metrics["doubles_real"]),
                             len(np.intersect1d(metrics["doubles_detected"],
                                                metrics["doubles_real"]))))

    print("Amplitude accuracy:")
    error = metrics["amps_real"][idxs_raw != -1] - \
            metrics["amps_extracted"][idxs_raw[idxs_raw != -1]]

    ax[0][1].set_title("amplitude error")
    ax[0][1].hist(error, 40)

    ax[0][0].set_title("negative_amplitudes error")
    ax[0][0].hist(metrics["amps_real"][np.where(idxs_raw == -1)[0]], 40)


    idxs_raw[idxs_raw==-1]

    range_ = (min(metrics["amps_real"].min(), 
                  metrics["amps_extracted"].min()),
              max(metrics["amps_real"].max(), 
                  metrics["amps_extracted"].max()))

    ax[1][0].set_title("real vs generated amps hists")
    ax[1][0].hist(metrics["amps_real"], 80, \
                  fc=(1,0,0,0.5), label="real", range=range_)
    ax[1][0].hist(metrics["amps_extracted"], 80, fc=(0,0,1,0.5),
            label="extracted", range=range_)
    ax[1][0].legend()

    range_ = (min(metrics["pos_real"].min(), 
                  metrics["pos_extracted"].min()),
              max(metrics["pos_real"].max(), 
                  metrics["pos_extracted"].max()))

    ax[3][0].set_title("real vs generated pos hists")
    ax[3][0].hist(metrics["pos_real"], 80, \
                  fc=(1,0,0,0.5), label="real", range=range_)
    ax[3][0].hist(metrics["pos_extracted"], 80, fc=(0,0,1,0.5),
            label="extracted", range=range_)
    ax[3][0].legend()


@graphics
def draw_real_frame(meta, metrics, idx, real=True, frame_l=15, frame_r=25):
    """Отрисовка кадра реального события. На графике также указываются
    амплитуды и времена реальных и выделенных событий.

    @meta - Метаданные df.
    @metrics - Посчитанные метрики. @note test_on_df должа быть выполнена с
    флагом extr_frames=True.
    @idx - Индекс кадра.
    @real - Если True - будет выведен кадр реального события, если False -
    кадр восстановленного события.
    @frame_l - Размер кадра в бинах слева.
    @frame_r -Размер кадра в бинах справа.
    """
    freq = 1/meta['params']['sample_freq']

    if real:
        data = metrics['frames_real'][idx]
        pos = metrics['pos_real'][idx]
    else:
        data = metrics['frames'][idx]
        pos = metrics['pos_extracted'][idx]

    ev_l_time = pos - frame_l*freq*1e+9
    ev_r_time = pos + frame_r*freq*1e+9

    l_index_r = np.searchsorted(metrics['pos_real'], ev_l_time)
    r_index_r = np.searchsorted(metrics['pos_real'], ev_r_time)

    l_index_e = np.searchsorted(metrics['pos_extracted'], ev_l_time)
    r_index_e = np.searchsorted(metrics['pos_extracted'], ev_r_time)

    poses_real = metrics['pos_real'][l_index_r: r_index_r] - ev_l_time
    poses_extr = metrics['pos_extracted'][l_index_e: r_index_e] - ev_l_time


    fig, ax = plt.subplots()
    ax.plot(data, label='frame data')
    ax.plot(poses_real/(1e+9*freq),
             metrics['amps_real'][l_index_r: r_index_r], 'ro',
             label='real events')
    ax.plot(poses_extr/(1e+9*freq),
             metrics['amps_extracted'][l_index_e: r_index_e], 'bo',
             label='extracted events')

    for i in range(l_index_r, r_index_r):
        if metrics['real_detected_transitions'][i] != -1:

            x = (metrics['pos_real'][i] - ev_l_time) / (1e+9*freq)
            y = metrics['amps_real'][i]

            trans_ind = metrics['real_detected_transitions'][i]
            trans_x = (metrics['pos_extracted'][trans_ind] - ev_l_time) / \
                      (1e+9*freq)
            trans_y = metrics['amps_extracted'][trans_ind]
            ax.arrow(x, y, trans_x - x, trans_y - y, fc='k', ec='k')


    ax.legend()


# -*- coding: utf-8 -*-
"""Скрипт предназначен для вычитания шумов из спектров.

Для определения шума берутся файлы с высоким запирающим напряжением.
На вход должны подаваться обработанные точки.

Input:
fill_dir - path to folder with processed Lan10-12PCI fill data
points_high - high voltage points wildcard
points_high_2 - high voltage points 2 wildcard
points_path - list of low HV points wildcard

Output:
Averaged energy spectrums with and w/o corrected tails for list of intresred
points.
"""
# pylint: disable-msg=R0914

from glob import glob
from os import path

import dfparser
import matplotlib.pyplot as plt
import numpy as np
import seaborn


def get_amps(filepath: str) -> np.ndarray:
    """Extract amplitudes from processed file."""
    _, _, data = dfparser.parse_from_file(filepath)
    p_high = dfparser.Point()
    p_high.ParseFromString(data)

    amps = np.hstack([list(block.events.amplitudes)
                      for block in p_high.channels[0].blocks])
    return amps


def _main():
    fill_dir = "/home/chernov/data_processed/2017_05/Fill_3/"
    points_high = "*/p1(30s)(HV1=18600).df"
    points_high_2 = "*/p2(30s)(HV1=18550).df"
    points_high_filepaths = glob(path.join(fill_dir, points_high)) + \
        glob(path.join(fill_dir, points_high_2))

    points_path = ["*/p0(30s)(HV1=16000).df",
                   "*/p36(30s)(HV1=17000).df",
                   "*/p80(30s)(HV1=15000).df",
                   "*/p102(30s)(HV1=14000).df"]

    hist_bins_num = 50
    hist_range = (700, 1500)

    amps_all = np.hstack([get_amps(fp) for fp in points_high_filepaths])
    hist_n, bins_n = np.histogram(amps_all, hist_bins_num, range=hist_range)
    hist_n = hist_n/len(points_high_filepaths)

    # plt.step((bins_n[1:] + bins_n[:-1])/2, hist_n)

    fig_all, ax_all = plt.subplots()
    ax_all.set_yscale("log", nonposx='clip')
    fig_all.canvas.set_window_title("Corrected points tails")
    fig_all.suptitle("Corrected points tails")
    ax_all.set_xlabel("Bins, ch")

    for point_path in points_path:
        points = glob(path.join(fill_dir, point_path))
        amps_p = np.hstack([get_amps(fp) for fp in points])
        hist_p, _ = np.histogram(amps_p, hist_bins_num, range=hist_range)
        hist_p = hist_p/len(points)
        hist_corr = hist_p - hist_n

        fig, axes = plt.subplots()
        fig.canvas.set_window_title(point_path)
        fig.suptitle("Tail with removed noise")
        axes.set_title("File - %s." % (point_path,))
        axes.set_xlabel("Bins, ch")
        axes.set_xlim(*hist_range)
        axes.set_yscale("log", nonposx='clip')
        axes.step((bins_n[1:] + bins_n[:-1])/2, hist_p, label="Original")
        axes.step((bins_n[1:] + bins_n[:-1])/2, hist_corr, label="Corrected")
        ax_all.step((bins_n[1:] + bins_n[:-1])/2, hist_corr, label=point_path)
        axes.legend()

    ax_all.legend()


if __name__ == "__main__":
    seaborn.set_context("poster")
    _main()



# -*- coding: utf-8 -*-
"""Определение плохих наборов.

Плохие сеты выделяются на основе отклонений по хи-квадрат от среднего по всем
сетам спектра.

Детектор работает только на преобразованных в события данных с Лан10-12PCI. (
Обработка производится скриптом ./scripts/convert_points.py)

Алгоритм работы:
1. Расчет энергетических спектров для каждой точки для всех сетов в выбранном
   наборе.
   Влияющие параметры:
   - ind-start - минимальный индекс точки в сете, которая будет рассмотрена
   - ind-end - максимальный индекс точки в сете, которая будет рассмотрена
   - ampl-threshold - порог по обрезке каналов снизу. Нижние каналы могут быть
     сильно зашумлены и в итоге плохо повлияют на работу алгоритма.
   - ampl-max - порог по обрезке каналов сверху
   - bins - количество бинов в энергетическом спектре. В силу параметров
     аппаратуры стоит подбирать количество бинов с расчетом на то, чтобы
     координаты краев бинов были кратны 16
   - norming_channels - каналы, по которым будет производится нормировка
     энергетических спектров. Имеет смысл захватывать пик гистограммы. Если
     параметр не указан - нормировка будет производится по интегралу всей
     гистограммы
2. Вычисление усредненных гистограмм для точек с одинаковыми индексами по всем
   сетам в наборе.
3. Вычисление хи-квадрат отклонений от энергетического спектра точки для каждой
   точки во всех сетах.
4. Вывод результата в виде графика.

# TODO: add time filtering (make as package tool?)
# TODO: сделать обработку norming_channels
"""
import glob
from argparse import ArgumentParser
from contextlib import closing
from functools import partial
from os import path

import dfparser
import numpy as np

import matplotlib.pyplot as plt
import seaborn as sns
from multiprocess import Pool
from natsort import natsorted


def __parse_args():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('data_root', help='Data root folder.')
    parser.add_argument('fill', help='Fill folder relative to data root.')
    parser.add_argument('--ind-start', type=int, default=0,
                        help='Points starting index (default - 0).')
    parser.add_argument('--ind-end', type=int, default=102,
                        help='Points ending index (default - 102).')
    parser.add_argument('-t', '--ampl-threshold', type=int, default=496,
                        help='Amplitudes threshold (default - 496).')
    parser.add_argument('-m', '--ampl-max', type=int, default=4016,
                        help='Amplitudes max threshold (default - 4016).')
    parser.add_argument('-b', '--bins', type=int, default=55,
                        help='Histogram bins number (default - 55).')
    return parser.parse_args()


def get_set_spectrum(set_abs_path, borders=None, bins=30):
    """Calculate energy spectrum for set."""
    points = natsorted(glob.glob(path.join(set_abs_path, "p*.df")))

    out = {}

    for point in points:
        _, meta, data = dfparser.parse_from_file(point)
        parsed_data = dfparser.Point()
        parsed_data.ParseFromString(data)
        del data

        amps = []
        times = []
        for channel in parsed_data.channels:
            for block in channel.blocks:
                amps.append(np.array(block.events.amplitudes, np.int16))
                times.append(np.array(block.events.times, np.uint64))

        amps = np.hstack(amps)
        times = np.hstack(times)
        hist, bins = np.histogram(amps, bins, range=borders, density=True)
        hist_unnorm, _ = np.histogram(amps, bins, range=borders)
        out[path.relpath(point, set_abs_path)] = {
            "meta": meta,
            "hist": hist,
            "hist_unnorm": hist_unnorm,
            "bins": bins
        }
    return out


def calc_hist_avg(sets_data, point_index):
    """Calculate point average histogram in sets."""
    sets = list(sets_data.values())
    hist_avg = np.zeros(ARGS.bins)
    used = 0
    for set_ in sets:
        for point in set_:
            curr_index = int(
                set_[point]["meta"]["external_meta"]["point_index"]
            )
            if curr_index == point_index:
                hist_avg += set_[point]["hist"]
                used += 1
    hist_avg /= used
    return hist_avg


def calc_chi_square(set_, hist_avg, point_index):
    """Calculate chi2 for set (unnormed).

    Parameters
    ----------
    set_ : Set data. Return value from get_set_spectrum().

    hist_avg: Average histogram data. Return value from calc_hist_avg().

    point_index: Index of point.

    Returns
    -------
    chi2: Unnormed chi2 for point.

    See Also
    --------
    get_set_spectrum

    """
    for point in set_:
        curr_index = int(
            set_[point]["meta"]["external_meta"]["point_index"]
        )
        if curr_index == point_index:
            hist = set_[point]["hist"]
            hist_unnorm = set_[point]["hist_unnorm"]
            sigma_raw = hist_unnorm ** 0.5
            sigma = sigma_raw * (hist / hist_unnorm)
            sigma[np.isnan(sigma)] = 1  # replace nans for empty bins
            return ((hist - hist_avg)**2 / sigma**2).sum() / ARGS.bins
    return None


def __calc_xticklabels(sets_data):
    labels = []
    for idx in range(ARGS.ind_start, ARGS.ind_end):
        for set_idx in sets_data:
            found = False
            for point in sets_data[set_idx]:
                ext_meta = sets_data[set_idx][point]["meta"]["external_meta"]
                curr_index = int(
                    ext_meta["point_index"]
                )
                if curr_index == idx:
                    found = True
                    hv = format(float(ext_meta["HV1_value"]) / 1000, '.2f')
                    labels.append("%s:%s" % (idx, hv))
                    break
            if found:
                break
    return labels


def __calc_point_sigma(sets_data, point_index):
    """Caluclate sigma based on fitting.

    NOTE: unfinished
    """
    sets = list(sets_data.values())
    hists = []
    for set_ in sets:
        for point in set_:
            curr_index = int(
                set_[point]["meta"]["external_meta"]["point_index"]
            )
            if curr_index == point_index:
                hists.append(np.arrayset_[point]["hist"])
    hists = np.vstack(hists)


def __main():
    group_abs = path.join(ARGS.data_root, ARGS.fill)
    sets = natsorted(glob.glob(path.join(group_abs, "set_*[0-9]")))[5:]
    get_spectrum = partial(get_set_spectrum, borders=(
        ARGS.ampl_threshold, ARGS.ampl_max), bins=ARGS.bins)

    with closing(Pool()) as pool:
        out = pool.map(get_spectrum, sets)

    sets_data = {path.relpath(s, group_abs): d for s, d in zip(sets, out)}

    hists_avg = []
    for idx in range(ARGS.ind_start, ARGS.ind_end):
        hists_avg.append(calc_hist_avg(sets_data, idx))

    sets_points_chi2 = {}
    for set_idx in sets_data:
        x_values = []
        y_values = []
        for idx in range(ARGS.ind_start, ARGS.ind_end):
            chi2 = calc_chi_square(sets_data[set_idx], hists_avg[idx], idx)
            if chi2:
                x_values.append(idx)
                y_values.append(chi2)
        sets_points_chi2[set_idx] = {"x": x_values, "y": y_values}

    _, axes = plt.subplots()

    axes.set_title(r'$\chi^2$ deviations for each point in %s' % (ARGS.fill) +
                   '\nBad sets excluded')
    axes.set_xlabel(r'Point index and HV, V')
    axes.set_ylabel(r'$\chi^2$')
    axes.set_yscale('log')
    axes.set_xticks(list(range(ARGS.ind_start, ARGS.ind_end)))
    axes.set_xticklabels(__calc_xticklabels(sets_data))
    for tick in axes.get_xticklabels():
        tick.set_rotation(90)
    palette = sns.color_palette("hls", len(sets_points_chi2))
    for idx, set_idx in enumerate(natsorted(sets_points_chi2.keys())):
        axes.scatter(
            sets_points_chi2[set_idx]["x"], sets_points_chi2[set_idx]["y"],
            label=set_idx, s=30, c=palette[idx], edgecolors="face"
        )
    axes.legend()
    plt.show()


if __name__ == "__main__":
    ARGS = __parse_args()
    sns.set_context("poster")
    sns.set_style(rc={"font.family": "monospace"})
    DEF_PALLETE = sns.color_palette()
    sns.set_palette(sns.cubehelix_palette(8))
    __main()



"""Correct set numbers between MADC and Lan10 datasets.

Script will:
- Correct set numbers and adds metadata to Lan10 sets.
  Detached from MADC sets will be marked.
- Compress points binary data.

"""
from functools import partial
from contextlib import closing
from multiprocessing import Pool
import glob
import shutil
import struct
import zlib
from argparse import ArgumentParser
from enum import Enum
from os import listdir, makedirs, path

import dfparser
from dateutil import parser as timeparser
from natsort import natsorted


def __parse_args():
    parser = ArgumentParser(description=__doc__)

    parser.add_argument('LAN10_PATH', help='Path to Lan10 dataset.')
    parser.add_argument('MADC_PATH', help='Path to MADC dataset.')
    parser.add_argument('OUT_PATH', help='Output data root path.')
    parser.add_argument('--err-sec', '-e', type=float, default=10.0,
                        help='Error in seconds between datasets.')
    parser.add_argument('-t', '--threads', type=int, default=None,
                        help='Processing threads (default - cpu cores count).')
    parser.add_argument('-o', '--overwrite', action='store_true',
                        help='Overwrite files in output directory.')

    return parser.parse_args()


def read_madc_sets(madc_sets_raw, group_abs_path):
    """Read begin and end times for each set in group.

    Parameters
    ----------
    madc_sets_raw : list
       Sets list for current group.
    group_abs_path : str
       Group absolute path.
    Returns
    -------
    madc_sets : dictionary
       A dictionary keyed by sets containing begin and end times.

    """
    madc_sets = {}
    for madc_set in madc_sets_raw:
        files = listdir(
            path.join(group_abs_path, madc_set))
        points = natsorted([f for f in files if f.startswith('p')])

        if points:
            _, meta_p_0, _ = dfparser.parse_from_file(
                path.join(group_abs_path, madc_set,
                          points[0]), nodata=True)

            _, meta_p_last, _ = dfparser.parse_from_file(
                path.join(group_abs_path, madc_set,
                          points[-1]), nodata=True)

            madc_sets[madc_set] = {
                "begin": timeparser.parse(meta_p_0["start_time"][0]),
                "end": timeparser.parse(meta_p_last["end_time"][-1])
            }

    return madc_sets


class SetState(Enum):
    """Set match state."""

    CORRUPTED = 1
    EMPTY = 2
    DETACHED = 3


def match_lan10_sets(lan10_sets_raw, lan10_root,
                     lan10_group_rel_path, madc_sets, err_sec):
    """Find correspondense between lan10 and madc sets.

    Parameters
    ----------
    lan10_sets_raw : list
       Lan10 sets list for current group.
    lan10_group_abs_path : str
       Lan10 group absolute path.
    madc_sets : str
       MADC sets time borders dictionary (return value from read_madc_sets).
    err_sec : float
       Maximal error between borders in seconds.
    Returns
    -------
    corrs : dictionary
       Sets mapping.

    """
    corrs = []
    for lan10_set in lan10_sets_raw:
        set_abs_path = path.join(lan10_root, lan10_group_rel_path, lan10_set)
        set_rel_path = path.relpath(set_abs_path, lan10_root)
        files = listdir(set_abs_path)
        points = natsorted([f for f in files if f.startswith('p')])
        if points:
            check_beginning = True
            try:
                _, meta, _ = dfparser.parse_from_file(
                    path.join(set_abs_path, points[0]), nodata=True)
                begin = timeparser.parse(meta["params"]["start_time"])
            except struct.error:
                check_beginning = False

            check_ending = True
            try:
                _, meta, _ = dfparser.parse_from_file(
                    path.join(set_abs_path, points[-1]), nodata=True)
                end = timeparser.parse(meta["params"]["end_time"])
            except struct.error:
                check_ending = False

            if not (check_beginning or check_ending):

                corrs.append({"from": set_rel_path, "to": SetState.CORRUPTED})
            else:
                detached = True
                for key in madc_sets:
                    if check_beginning:
                        begin_err = abs(
                            (madc_sets[key]["begin"] - begin)
                            .total_seconds()
                        )
                    if check_ending:
                        end_err = abs(
                            (madc_sets[key]["end"] -
                             end).total_seconds()
                        )
                    if (not check_beginning or begin_err < err_sec) and \
                            (not check_ending or end_err < err_sec):

                        corrs.append({
                            "from": set_rel_path,
                            "to": path.join(path.dirname(set_rel_path), key)
                        })
                        detached = False
                        break

                if detached:
                    corrs.append({
                        "from": set_rel_path,
                        "to": SetState.DETACHED
                    })
        else:
            corrs.append({
                "from": set_rel_path,
                "to": SetState.EMPTY
            })
    return corrs


def process_set(lan10_root, madc_root, out_root, overwrite, corrs):
    """Process group dataset.

    Parameters
    ----------
    lan10_root : str
       Lan10 data root path.
    out_root : str
       Output data root path.
    corrs : list
       Correspondency list between sets (return value from match_lan10_sets).

    """
    for corr in corrs:
        if corr["to"] == SetState.CORRUPTED:
            out_path = path.join(out_root, "%s-corrupted" % corr["from"])
            if path.exists(out_path):
                if overwrite:
                    shutil.rmtree(out_path)
                else:
                    continue
            shutil.copytree(path.join(lan10_root, corr["from"]), out_path)

        elif corr["to"] == SetState.DETACHED:
            out_path = path.join(out_root, "%s-detached" % corr["from"])
            if path.exists(out_path):
                if overwrite:
                    shutil.rmtree(out_path)
                else:
                    continue
            shutil.copytree(path.join(lan10_root, corr["from"]), out_path)

        elif corr["to"] == SetState.EMPTY:
            out_path = path.join(out_root, "%s-empty" % corr["from"])
            if path.exists(out_path):
                if overwrite:
                    shutil.rmtree(out_path)
                else:
                    continue
            shutil.copytree(path.join(lan10_root, corr["from"]), out_path)

        else:
            lan10_set_path = path.join(lan10_root, corr["from"])
            madc_set_path = path.join(madc_root, corr["to"])
            out_set_path = path.join(out_root, corr["to"])
            if path.exists(out_set_path):
                if overwrite:
                    shutil.rmtree(out_set_path)
                else:
                    continue

            makedirs(out_set_path)

            for metafile in ["meta", "voltage", "scenario"]:
                if path.exists(path.join(madc_set_path, metafile)):
                    shutil.copy(path.join(madc_set_path, metafile),
                                path.join(out_set_path, metafile))

            for point in listdir(lan10_set_path):
                try:
                    _, meta, data = dfparser.parse_from_file(
                        path.join(lan10_set_path, point)
                    )
                    meta["compression"] = "zlib"
                    with open(path.join(out_set_path, point), "wb") as out_file:
                        out_file.write(
                            dfparser.create_message(
                                meta, zlib.compress(data)
                            )
                        )
                except struct.error:
                    shutil.copy(
                        path.join(lan10_set_path, point),
                        path.join(out_set_path, "%s-corrupted" % point)
                    )


def process_group(group, args):
    """Process all sets in group."""
    madc_sets_raw = [
        f for f in listdir(path.join(args.MADC_PATH, group))
        if f.startswith("set_")]
    lan10_sets_raw = [
        f for f in listdir(path.join(args.LAN10_PATH, group))
        if f.startswith("set_")]

    madc_sets = read_madc_sets(
        madc_sets_raw, path.join(args.MADC_PATH, group)
    )

    corrs = match_lan10_sets(
        lan10_sets_raw, args.LAN10_PATH, group, madc_sets, args.err_sec
    )

    process_set(args.LAN10_PATH, args.MADC_PATH, args.OUT_PATH,
                args.overwrite, corrs)


def __main():
    args = __parse_args()

    lan10_groups = [path.relpath(g, args.LAN10_PATH) for g in list(
        set([path.dirname(s) for s in glob.glob(
            path.join(args.LAN10_PATH, "**", "set_*"), recursive=True)])
    )]

    madc_groups = [path.relpath(g, args.MADC_PATH) for g in list(
        set([path.dirname(s) for s in glob.glob(
            path.join(args.MADC_PATH, "**", "set_*"), recursive=True)])
    )]

    merging_groups = [
        grp for grp in lan10_groups if grp in madc_groups]

    process_group_part = partial(process_group, args=args)
    with closing(Pool(args.threads)) as pool:
        pool.map(process_group_part, merging_groups)


if __name__ == "__main__":
    __main()



"""Data files conversion utils."""
# pylint: disable-msg=C0413

from os import path
import sys

from dateutil.parser import parse
import dfparser
import numpy as np

CUR_DIR = path.dirname(path.realpath(__file__))
if CUR_DIR not in sys.path:
    sys.path.append(CUR_DIR)
del CUR_DIR

from process_utils import extract_frames


def apply_zsupression(data: np.ndarray, threshold: int=500, area_l: int=50,
                      area_r: int=100) -> tuple:
    """Обрезание шумов в файле данных платы Лан10-12PCI.

    Функция расчитана на файлы данных с максимальным размером кадра
    (непрерывное считывание с платы).

    @data - данные кадра (отдельный канал)
    @threshold - порог амплитуды события
    @area_l - область около события, которая будет сохранена
    @area_r - область около события, которая будет сохранена

    @return список границ события

    """
    peaks = np.where(data > threshold)[0]
    dists = peaks[1:] - peaks[:-1]
    gaps = np.append(np.array([0]), np.where(dists > area_r)[0] + 1)

    events = ((peaks[gaps[gap]] - area_l, peaks[gaps[gap + 1] - 1] + area_r)
              for gap in range(0, len(gaps) - 1))

    return events


def rsb_to_df(ext_meta: dict, rsb_file, threshold: int=500,
              area_l: int=50, area_r: int=100) -> (dict, bytearray, int):
    """Конвертировние данных формата rsb в формат df.

    @meta - метаданные сообщения с точками
    @rsb_file - файл с платы Руднева-Шиляева
    @threshold - порог амплитуды события (параметр zero-suppression)
    @area_l - область около события, которая будет сохранена (параметр
    zero-suppression)
    @area_r - область около события, которая будет сохранена (параметр
    zero-suppression)

    @return - (meta, data, data_type)

    """
    sec_coef = 1e+9

    rsb_ds = dfparser.RshPackage(rsb_file)

    meta = {}
    meta["external_meta"] = ext_meta
    meta["params"] = rsb_ds.params
    meta["process_params"] = {
        "threshold": threshold,
        "area_l": area_l,
        "area_r": area_r
    }

    begin_time = parse(rsb_ds.params["start_time"]).timestamp()*sec_coef
    end_time = parse(rsb_ds.params["end_time"]).timestamp()*sec_coef
    bin_time = (rsb_ds.params["sample_freq"]**-1)*sec_coef
    b_size = rsb_ds.params["b_size"]

    if rsb_ds.params["events_num"] == -1:
        meta["recalc_events_num"] = True
        rsb_ds.params["events_num"] = np.iinfo(int).max
        for i in range(np.iinfo(int).max):
            try:
                rsb_ds.get_event(i)
            except Exception as exception:
                rsb_ds.params["events_num"] = i
                break

    events_num = rsb_ds.params["events_num"]
    ch_num = rsb_ds.params["channel_number"]

    use_time_corr = False
    if events_num > 0:
        event = rsb_ds.get_event(0)
        if "ns_since_epoch" not in event:
            use_time_corr = True
            times = list(np.linspace(begin_time, end_time -
                                     int(bin_time*b_size),
                                     events_num))
            meta["correcting_time"] = "linear"

    point = dfparser.Point()
    channels = [point.channels.add(id=channel) for channel in range(ch_num)]
    for i in range(events_num):
        event_data = rsb_ds.get_event(i)

        if use_time_corr:
            time = times[i]
        else:
            time = event_data["ns_since_epoch"]

        for channel in range(ch_num):
            block = channels[channel].blocks.add(time=int(time), )

            ch_data = event_data["data"][channel::ch_num]
            for frame in apply_zsupression(ch_data, threshold, area_l, area_r):
                frame = np.clip(frame, 0, ch_data.shape[0] - 1)
                event = block.frames.add()
                event.time = int(frame[0]*bin_time)
                event.data = ch_data[frame[0]:frame[1]].astype(np.int16) \
                    .tobytes()

    meta["bin_offset"] = 0
    meta["bin_size"] = point.ByteSize()
    data = point.SerializeToString()

    return meta, data


def df_frames_to_events(meta, data, extract_func, frame_l=15, frame_r=25,
                        correct_time=False):
    """Convert frames to events in dataforge points."""
    threshold = meta['process_params']['threshold']
    sample_freq = meta['params']['sample_freq']

    point = dfparser.Point()
    point.ParseFromString(data)

    for channel in point.channels:
        for block in channel.blocks:

            events = block.events
            for _ in range(len(block.frames)):
                frame = block.frames.pop(0)

                if correct_time:
                    frame.time = (frame.time//np.uint64(10)).astype(np.uint64)

                ev_data = np.frombuffer(frame.data, np.int16)
                params, singles_raw = extract_func(ev_data, frame.time,
                                                   threshold, sample_freq)

                singles = np.where(singles_raw == True)[0]

                events.times.extend(np.round(params[singles*2 + 1])
                                    .astype(np.uint64))
                events.amplitudes.extend(np.round(params[singles*2])
                                         .astype(np.uint64))

                doubles = np.where(singles_raw == False)[0]
                events_bins = np.round((params[doubles*2 + 1] -
                                        frame.time) * sample_freq / 1e+9)

                frames_block = extract_frames(ev_data, events_bins,
                                              frame_l, frame_r)
                for idx, frame in enumerate(frames_block):
                    time = int(np.round(params[doubles[idx]*2 + 1]) +
                               frame_l / sample_freq * 1e+9)
                    frame_ser = frame.astype(np.int16).tobytes()
                    block.frames.add(time=time, data=frame_ser)

    meta['process_params']['extracted'] = True
    meta['process_params']['extracted_frame_l'] = frame_l
    meta['process_params']['extracted_frame_r'] = frame_r

    return meta, point.SerializeToString()



"""Install script."""

from pip.req import parse_requirements
from setuptools import find_packages, setup

INSTALL_REQS = parse_requirements("signal_utils/requirements.txt",
                                  session='hack')
REQS = [str(ir.req) for ir in INSTALL_REQS]

setup(
    name="signal_utils",
    description='',
    version="0.1.21",
    author="Vasiliy Chernov",
    packages=find_packages(),
    platforms='any',
    install_requires=REQS,
    include_package_data=True,
    package_data={
        '': ['*.h5', '*.dat'],
    }
)



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 27 19:49:34 2017

@author: kapot
"""

import io
from datetime import datetime

import numpy as np
from dfparser import dump_to_rsb
from dfparser.df_data.def_values import DEF_RSH_PARAMS
from random_custom_pdf import rand_custom
from scipy.interpolate import interp1d
from tqdm import tqdm

from convert_utils import rsb_to_df


def generate_noise(x, base_freq=3125000.0):
    """
      @x - координаты в секундах
      @base_freq - частота, при которой подбирались параметры сигнала
      @return - шум для заданных координат

    """
    x_min = int(x.min() * base_freq - 2)
    x_max = int(x.max() * base_freq + 2)

    x_orig = np.arange(x_min, x_max)

    noise = np.random.randint(-20, 20, size=len(x_orig))
    c1 = np.convolve(noise, np.full((10), 0.315), 'same')
    c2 = np.convolve(c1, np.full((3), 1 / 3), 'same')
    c3 = np.convolve(c2, np.full((3), 1 / 3), 'same') + 2.5

    noise_conv = np.round(c3)

    interp = interp1d(x_orig / base_freq, noise_conv, kind='nearest')

    return interp(x) * 16


def gen_signal(x, ampl, pos,
               sigma=0.34156352,
               tail_amp=0.37015974,
               tail_factor=2.96038975,
               p=2.20563286,
               s=0.0525933,
               base_freq=3125000.0):
    """
      Генерация сигнала

      Все аргументы, кроме @ampl и @pos подобраны под реальный сигнал.
      Рекомендуется оставить из значения по умолчанию

      @x - координаты в секундах
      @ampl - амплитуда генерируемого сигнала
      @pos - положение генерируемого сигнала в секундах
      @sigma - растяжение сигнала
      @tail_amp - отношение амлитуды выброса к амплитуде сигнала
      @tail_factor - степень резкости пика сигнала
      @p - степень гауссиана амплитуда
      @s - коэффициент растяжения выброса
      @base_freq - частота, при которой подбирались параметры сигнала
      @return - сигнал для заданных координат

    """

    def gauss(x): return np.exp(
        (-1 / 2) * np.power((np.abs(sigma * x * base_freq)), p))

    def gauss_rev(y): return (
        (1 / sigma) * np.power(-2 * np.log(y), 1 / p) / base_freq)

    def spike(x): return (1 / (1 + 2 * x * base_freq * s)**tail_factor - 1.0) * \
        np.exp(-x * base_freq * s)

    spike_offset = gauss_rev(0.1) + pos
    spike_x = x - spike_offset
    spike_x[spike_x < 0] = 0

    return ((gauss(x - pos) + spike(spike_x) * tail_amp) * ampl)


def gen_multiple(x, *args, l_size=10 / 3125000.0, r_size=100 / 3125000.0):
    """
      Функция нескольких событий. Для ускорения работы генерируемый сигнал
      обрезается после l_size от пика слева и на r_size от пика справа.

      @x - упорядоченная сетка
      @a - амплитуда текущего события
      @p - положение пика текущего события
      @args - последовательно указанные амплитуды
      @l_size - граница события слева
      @r_size - граница события справа
      и положения пиков остальных событий

    """
    assert(not len(args) % 2)

    y = np.zeros(len(x), np.float32)

    l_vals = np.searchsorted(x, [args[i + 1] - l_size for
                                 i in range(0, len(args), 2)])
    r_vals = np.searchsorted(x, [args[i + 1] + r_size for
                                 i in range(0, len(args), 2)])

    for i in range(0, len(args), 2):
        n = i // 2
        y[l_vals[n]:r_vals[n]] += gen_signal(x[l_vals[n]:r_vals[n]],
                                             args[i], args[i + 1])

    return y


def gen_raw_block(freq: float=12e+3,
                  sample_freq: float=3125000.0,
                  b_size: int=1048576,
                  dist_file: str=None,
                  dist_time_func=None,
                  min_amp: int=500,
                  max_amp: int=7000):
    """
      Генерация блока данных для тестирования
      @freq - частота событий (на секунду)
      @sample_freq - частота оцифровки Гц
      @b_size - размер блока в бинах
      @dist_file - файл с гистограммой распределения (см data/dist.dat).
      Перекрывает параметры min_amp и max_amp.
      @dist_time_func - Функция, генерирующая времена событий
      (event_num: int, max_time: float) -> np.ndarray(np.float128).
      Если не указано, будет использовано равномерное распределение.
      @max_amp - максимальная амплитуда события
      @min_amp - минимальная амплитуда события
      @return - [data, [amp1, pos1, amp2, pos2, ...]]

    """
    events = int(freq * (b_size / sample_freq))

    params = np.zeros(events * 2, np.float128)

    if dist_time_func is None:
        params[1::2] = np.sort(np.random.uniform(
            0,  b_size / sample_freq, events))
    else:
        params[1::2] = dist_time_func(events, b_size / sample_freq)\
            .astype(np.float128)

    if not dist_file:
        params[0::2] = np.random.uniform(min_amp, max_amp, events)
    else:
        if gen_raw_block.file != dist_file:
            data = np.fromfile(dist_file, sep='\t')
            data = data.reshape((len(data) // 2, 2))
            gen_raw_block.file = dist_file
            gen_raw_block.x = data[:, 0]
            gen_raw_block.y = data[:, 1]

        params[0::2] = rand_custom(gen_raw_block.x, gen_raw_block.y, (events,))

    x = np.arange(b_size) / sample_freq
    data = gen_multiple(x, *params)

    return data + generate_noise(x), params


gen_raw_block.file = None
gen_raw_block.x = None
gen_raw_block.y = None


def generate_multiple_blocks(freq: float=12e+3,
                             sample_freq: float=3125000.0,
                             b_size: int=1048576,
                             dist_file: str=None,
                             dist_time_func=None,
                             min_amp: int=500,
                             max_amp: int=7000,
                             time: float=30.0,
                             block_write_time: float=0.05):
    """
        Генерация множественных блоков

        @freq - частота событий (на секунду)
        @sample_freq - частота оцифровки Гц
        @b_size - размер блока в бинах
        @dist_file - файл с гистограммой распределения (см data/dist.dat).
        Перекрывает параметры min_amp и max_amp.
        @dist_time_func - Функция, генерирующая времена событий
        (event_num: int) -> np.ndarray(np.float128). Если не указано, будет
        использовано равномерное распределение.
        @max_amp - максимальная амплитуда события
        @min_amp - минимальная амплитуда события
        @time - время набора в секундах
        @block_write_time - cкорость записи блока; определяет мертвое время
        (http://elog.mass.inr.ru/nu-mass/319)
        @return - [data, [amp1, pos1, amp2, pos2, ...], start_times];
        здесь data - данные блоков (block_num, block_size)
        [amp1, pos1, amp2, pos2, ...] - массив событий:
            amp1 - амплитуда события в каналах
            pos1 - абсолютное положение события в наносекундах с начала первого
            блока
            start_times - массив в наносекундах стартовых времен для
            каждого блока

    """
    block_time = ((b_size / sample_freq) + block_write_time)
    blocks_num = round(time / block_time + 0.5)

    start_times = np.array([block_time * i for i in range(blocks_num)])
    block_params = []

    data = np.zeros((blocks_num, b_size), np.int16)

    for i in tqdm(range(blocks_num), desc="generating blocks"):
        out = gen_raw_block(freq, sample_freq, b_size,
                            dist_file, dist_time_func, min_amp, max_amp)
        data[i] = out[0]
        out[1][1::2] += start_times[i]
        out[1][1::2] *= 1e+9

        block_params.append(out[1])

    block_params = np.hstack(block_params)

    return data, block_params, start_times * 1e+9


def generate_rsb(freq: float=12e+3,
                 sample_freq: float=3125000.0,
                 b_size: int=1048576,
                 dist_file: str=None,
                 dist_time_func=None,
                 min_amp: int=500,
                 max_amp: int=7000,
                 time: float=30.0,
                 block_write_time: float=0.05):
    """
      Генерация файла rsb с желаемой частотой событий

      @freq - частота событий (на секунду)
      @sample_freq - частота оцифровки Гц
      @b_size - размер блока в бинах
      @dist_file - файл с гистограммой распределения (см data/dist.dat).
      Перекрывает параметры min_amp и max_amp.
      @dist_time_func - Функция, генерирующая времена событий
      (event_num: int) -> np.ndarray(np.float128). Если не указано, будет
      использовано равномерное распределение.
      @max_amp - максимальная амплитуда события
      @min_amp - минимальная амплитуда события
      @time - время набора в секундах
      @block_write_time - cкорость записи блока; определяет мертвое время
      (http://elog.mass.inr.ru/nu-mass/319)
      @return - [file_data, [amp1, pos1, amp2, pos2, ...]];
      здесь file_data - данные сгенерированного rsb файла
      [amp1, pos1, amp2, pos2, ...] - массив событий:
            amp1 - амплитуда события в каналах
            pos1 - абсолютное положение события в наносекундах с начала первого
            блока

    """

    data, block_params, \
        start_times = generate_multiple_blocks(freq, sample_freq, b_size,
                                               dist_file, dist_time_func,
                                               min_amp, max_amp, time,
                                               block_write_time)

    time_start = datetime.now().timestamp() * 1e+9

    rsh_params = DEF_RSH_PARAMS
    rsh_params['aquisition_time'] = time * 1000
    rsh_params['sample_freq'] = sample_freq

    rsb = dump_to_rsb(rsh_params,
                      time_start + start_times,
                      data)
    block_params[1::2] += time_start

    return rsb, block_params


def generate_df(threshold: int=500,
                area_l: int=50,
                area_r: int=100,
                freq: float=12e+3,
                sample_freq: float=3125000.0,
                b_size: int=1048576,
                dist_file: str=None,
                dist_time_func=None,
                min_amp: int=500,
                max_amp: int=7000,
                time: float=30.0,
                block_write_time: float=0.05):
    """
      Генерация файла формата df

      @freq - частота событий (на секунду)
      @sample_freq - частота оцифровки Гц
      @b_size - размер блока в бинах
      @dist_file - файл с гистограммой распределения (см data/dist.dat).
      Перекрывает параметры min_amp и max_amp.
      @dist_time_func - Функция, генерирующая времена событий
      (event_num: int) -> np.ndarray(np.float128). Если не указано, будет
      использовано равномерное распределение.
      @max_amp - максимальная амплитуда события
      @min_amp - минимальная амплитуда события
      @time - время набора в секундах
      @block_write_time - cкорость записи блока; определяет мертвое время
      (http://elog.mass.inr.ru/nu-mass/319)
      @return - [meta, data, [amp1, pos1, amp2, pos2, ...]];
      Здесь:
          meta - метаданные сгенерированного df файла
          data - бинарные данные сгенерированного df файла
          [amp1, pos1, amp2, pos2, ...] - массив событий:
            amp1 - амплитуда события в каналах
            pos1 - абсолютное положение события в наносекундах с начала первого
            блока

    """

    rsb, block_params = generate_rsb(freq, sample_freq, b_size, dist_file,
                                     dist_time_func, min_amp,  max_amp, time,
                                     block_write_time)

    file = io.BytesIO(rsb)
    meta, data = rsb_to_df({'simulation': True,
                            'freq': freq,
                            'min_amp': min_amp,
                            'max_amp': max_amp,
                            'block_write_time': block_write_time},
                           file, threshold,
                           area_l, area_r)

    return meta, data, block_params



# TODO: remove hardcode
from glob import glob
from os import makedirs
from os import path
import shutil

lan10_root = "/data/lan10"
lan10_processed_root = "/data/lan10_processed/"

metas = [
    path.relpath(p, lan10_root) for p in
    glob(path.join(lan10_root, "**", "meta"), recursive=True)]
    
scenarios = [
    path.relpath(p, lan10_root) for p in
    glob(path.join(lan10_root, "**", "scenario"), recursive=True)]
    
voltages = [
    path.relpath(p, lan10_root) for p in
    glob(path.join(lan10_root, "**", "voltage"), recursive=True)]
    
for metafile in [*metas, *scenarios, *voltages]:
    src = path.join(lan10_root, metafile)
    dst = path.dirname(path.join(lan10_processed_root, metafile))
    if not path.exists(dst):
        makedirs(dst)
    shutil.copy(src, dst)
